#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>

#include "scanner.h"
#include "tree.h"
#include "nametab.h"
#include "genrtl.h"
#include "rtl.h"
#include "comp.h"
#include "genasm.h"

/* Globals */
char input_file_name[PATH_MAX + 2] = "";
char dump_file_name[PATH_MAX + 4] = "";
FILE *dump_file = NULL;

int compiler_flags = 0;

type base_type;
type base_parameter_type;
type_ptr current_type = &base_type;

name_ptr current_function = NULL;
int frame_size = 0;
int function_end_label = 0;
int continue_label = 0, break_label = 0;
int extern_def = 0;
int error_level = 0;

void set_type(type_ptr t, int lex_value) {
    if (lex_value == INT) {
        t->base_type = BASETYPE_INT;
        t->element_size = 4;
    } else if (lex_value == CHAR) {
        t->base_type = BASETYPE_CHAR;
        t->element_size = 1;
    } else if (lex_value == VOID) {
        t->base_type = BASETYPE_VOID;
        t->element_size = 1;
    }
    t->size = current_type->element_size;
    t->base_type_size = current_type->element_size;
}

void assign_type(name_ptr n, type_ptr t) {
    n->typ.base_type = t->base_type;
    n->typ.base_type_size = t->base_type_size;
    n->typ.element_size = t->element_size;
    if ((n->typ.decl & DECL_PNT) != 0)
        n->typ.element_size = 4;
    n->typ.size = n->typ.element_size;
    if (n->typ.dimension != 0)
        n->typ.size *= n->typ.dimension;
}

name_ptr process_variable(const char *name) {
    name_ptr n = name_lookup(name, current_table);

    /* printf("process_variable %s, name table = %p\n", name, current_table); */

    if (n != NULL) {
        error_message1("Multiple definition of", name);
        n = NULL;
    }
    if (n == NULL) {
        n = allocate_name_entry();
        strcpy(n->name, name);
        n->typ.decl = DECL_VAR;
        if (extern_def)
            n->modifiers |= (MOD_EXTERN | MOD_GLOBAL);
        assign_type(n, current_type);
        add_name_entry(current_table, n);
    }
    return n;
}

name_ptr process_function_name(const char *name) {
    name_ptr n = name_lookup(name, current_table);
    if (n != NULL && (n->typ.decl & DECL_FUN) == 0) {
        error_message1("Multiple definition of", name);
        n = NULL;
    }
    if (n == NULL) {
        n = allocate_name_entry();
        strcpy(n->name, name);
        n->typ.decl = DECL_FUN;
        /* In our language, we have no static functions. */
        /* So its name is global */
        n->modifiers |= MOD_GLOBAL;
        assign_type(n, current_type);
        add_name_entry(current_table, n);
    } else {
        /* At the moment, we don't check the arguments of function */
        /* Release the list of arguments defined previously */
        n->typ.num_arguments = 0;
        release_name_table(&(n->typ.arguments));
    }
    current_function = n;
    current_type = &base_parameter_type;
    current_table = &(n->typ.arguments);
    return n;
}

name_ptr process_function_call(const char *name) {
    name_ptr n = name_lookup(name, &local_table);
    if (n != NULL && (n->typ.decl & DECL_FUN) == 0) {
        error_message1("Call of non-function", name);
        n = NULL;
    }
    if (n == NULL) {
        n = name_lookup(name, &global_table);
        if (n != NULL && (n->typ.decl & DECL_FUN) == 0) {
            error_message1("Call of non-function", name);
            n = NULL;
        }
    }
    if (n == NULL) {
        /* Implicit function definition */
        n = allocate_name_entry();
        strcpy(n->name, name);
        n->typ.decl = DECL_FUN;
        n->modifiers |= MOD_GLOBAL;
        set_type(&n->typ, INT);
        add_name_entry(current_table, n);
    } else {
        /* Function is defined before */
        /* At the moment, we don't check the arguments of function */
    }
    return n;
}

/* Copy name to local table */
static int argnum = 0;
static int copy_to_local(block_ptr a, void *arg0, long arg1) {
    name_ptr n = allocate_name_entry();
    copy_name_entry(n, (name_ptr) a);
    n->arg_number = argnum--;
    lst_addAfter(&local_table, (block_ptr) n);
    return 0;
}

void copy_params_to_local_table(name_ptr f) {
    argnum = f->typ.num_arguments;
    lst_loopReverse(
        &(f->typ.arguments), &(f->typ.arguments),
        &copy_to_local, NULL, 0
    );
}

static int local_offset = (-4);
static int process_local_name(block_ptr b, void *arg0, long arg1) {
    int s;
    name_ptr n = (name_ptr) b;
    if (n->typ.decl & DECL_FUN)
        return 0;
    if (n->arg_number > 0) {
        n->offset = 4 * n->arg_number + 4;  /* for first arg, offset = 8 */
    } else {
        s = n->typ.size;
        assert(s > 0);
        if (s % 4 != 0) s += (4 - s % 4);   /* Align to word boundary */
        local_offset -= s;
        n->offset = local_offset + 4;
    }
    n->modifiers |= MOD_LOCAL;
    return 0;
}

/* Calculate offsets of local variables in a stack frame */
void process_local_table() {
    local_offset = (-4);
    lst_loop(&local_table, &process_local_name, NULL, 0);

    print_name_table(&local_table);
}

void gen_function_beg() {
    function_end_label = gen_getlabel();
    gen_start_function();       /* Initialize RTL-generator */
    add_rtl_function_beg();
    if (current_function->modifiers & MOD_FUNCT_BODY_DEFINED) {
        error_message1(
            "Multiple definition of function", current_function->name);
    }
    current_function->modifiers |= MOD_FUNCT_BODY_DEFINED;
}

void gen_function_end() {
    gen_label(function_end_label);
    add_rtl_function_end();

    print_rtl_function();   /* Dump RTL representation of function */

    /* Optimization, Assembler generation */
    genasm_generate_function(); 
}

void parse_command_line(int argc, char *argv[]) {
    int argIdx = 1;
    while (argIdx < argc && argv[argIdx] != NULL) {
        const char *arg = argv[argIdx];
        char c = arg[0];
        char c1 = 0, c2 = 0;
        if (c != 0)
            c1 = arg[1];
        if (c1 != 0)
            c2 = arg[2];
        if (c != '-') {
            /* This is file name */
            if (input_file_name[0] != 0) {
                print_help(); exit(1);  /* Error in command line */
            }
            strncpy(input_file_name, arg, PATH_MAX);
        } else if (c1 == 'd') {
            if (c2 == 0) {
                print_help(); exit(1);  /* Error in command line */
            }
            if (c2 == 'n') {
                compiler_flags |= DUMP_NAMETABLES;
            } else if (c2 == 't') {
                compiler_flags |= DUMP_TREES;
            } else if (c2 == 'r') {
                compiler_flags |= DUMP_RTL;
            } else if (c2 == 'f') {
                const char *dumpfile = NULL;
                if (arg[3] != 0) {
                    dumpfile = arg + 3;
                } else if (argIdx < argc -1) {
                    argIdx++;
                    dumpfile = argv[argIdx];
                }
                if (dumpfile == NULL || dumpfile[0] == '-') {
                    print_help(); exit(1);  /* Error in command line */
                }
                strncpy(dump_file_name, dumpfile, PATH_MAX);
            }
        } else if (c1 == 'o') {
            const char *outfile = NULL;
            if (arg[2] != 0) {
                outfile = arg + 2;
            } else if (argIdx < argc -1) {
                argIdx++;
                outfile = argv[argIdx];
            }
            if (outfile == NULL || outfile[0] == '-') {
                print_help(); exit(1);  /* Error in command line */
            }
            strncpy(output_file_name, arg, PATH_MAX);
        } else {
            print_help(); exit(1);  /* Unknown switch */
        }
        argIdx++;
    }

    if (input_file_name[0] != 0) {
        yyin = fopen(input_file_name, "r");
        if (yyin == NULL) {
            perror("Could not open input file"); exit(1);
        }
        if (output_file_name[0] == 0) {
            char *p;
            strcpy(output_file_name, input_file_name);
            /* Find extension */
            p = output_file_name + strlen(output_file_name);
            while (
                p > output_file_name &&
                *(p-1) != '/' && *(p-1) != '\\' &&
                *(p-1) != '.'
            )
                p--;
            if (p > output_file_name && *(p-1) == '.') {
                *p = 's'; *(p+1) = 0;
            } else {
                strcat(output_file_name, ".s");
            }
        }
    } else {
        yyin = stdin;
    }

    if (dump_file_name[0] != 0) {
        dump_file = fopen(dump_file_name, "w");
        if (dump_file == NULL) {
            perror("Could not open the dump file"); exit(1);
        }
    } else {
        dump_file = stdout;
    }
}

void print_help() {
    printf(
        "The model c-compiler.\n"
        "Usage:\n"
        "    tinycc [-dn] [-dt] [-dr] [-df dumpFile] [-o asmFile] [inputFile]\n"
        "where\n"
        "    -dn         dump name tables,\n"
        "    -dt         dump trees,\n"
        "    -dr         dump RTL representation,\n"
        "    -df file    dump file path,\n"
        "    -o file     Assembler output file name\n"
        "                (default is input file name with \".s\" extension\n"
        "                or \"out.s\", if input file is not set),\n"
    );
}

void error_message0(const char *txt) {
    fprintf(stderr, "Error near line %d: %s.\n", yylineno, txt);
    error_level++;
}

void error_message1(const char *txt, const char *arg) {
    fprintf(stderr, "Error near line %d: %s %s.\n", yylineno, txt, arg);
    error_level++;
}
