/*
 * Generation of Assembler output
 */
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include "genrtl.h"
#include "rtl.h"
#include "nametab.h"
#include "scanner.h"
#include "strpool.h"
#include "comp.h"
#include "genasm.h"
#include "target.h"

char output_file_name[PATH_MAX + 2] = "";
FILE *output_file = NULL;

int *pseudo_reg_assignment = NULL;
int *pseudo_reg_first_usage = NULL;
int *pseudo_reg_last_usage = NULL;
int *pseudo_register_map = NULL;
int *pseudo_register_stack_slot = NULL;
int pseudo_reg_array_size = 0;

int stack_32_slots = 0;         /* Map of first 32 stack slots */
int stack_additional_space = 0; /* In words, including first 32 slots */
int stack_max_additional_space = 0; /* max value of stack_additional_space */
int stack_slots_offset = 0;     /* Offset from FP to bottom of stack slots */

int stack_save_hardregs_depth = 0;
int stack_save_hardregs_maxdepth = 0;

/* Set of clobbered registers */
int clobbered_registers = 0;

/* Set of currently allocated registers */
int allocated_registers = 0;

int genasm_pass = 0;    /* Pass counter */

/* Basic block */
int basic_block_begin = 0; /* BB Beginning: Line number of "goto", "label", etc. */
int basic_block_end = 0;   /* BB End ("goto", "label", etc.) */

/* Static functions prototypes */
static void scan_register_usage_rtx(rtx_ptr expr, int line_number, int level);
static int scan_rtl_line(block_ptr line, void *line_number, long arg1);
static int generate_rtl_line(block_ptr line, void *line_number, long arg1);
static void generate_rtl_instr(rtl_ptr line, int line_number);
static void generate_rtl_goto(rtl_ptr line);
static void generate_rtl_label(rtl_ptr line);
static void allocate_reg(int pseudoreg);
static void release_regs_rtx(rtx_ptr expr, int line_number);
static void allocate_regs_rtx(rtx_ptr expr, int line_number);
static int allocate_regs_rtl_line(
    block_ptr line, void *line_num, long arg1
);
static void generate_save_hreg(int hreg);
static void generate_restore_hreg(int hreg);
static void generate_push_hreg(int hreg);
static void generate_pop_hreg(int hreg);
static void generate_move_hregs(int hreg_src, int hreg_dst);
static void generate_move_pseudoreg_to_hreg(int pseudoreg, int hreg);
static void generate_move_hreg_to_pseudoreg(int hreg, int pseudoreg);
static instr_def_cptr match_instr(rtx_ptr expr);
static int is_reg_mem(rtx_ptr expr);
static int is_reg_hreg(rtx_ptr expr);
static int is_reg(rtx_ptr expr);
static int is_hreg(rtx_ptr expr);
static int is_mem(rtx_ptr expr);
static int is_bytemem(rtx_ptr expr);
static int is_const(rtx_ptr expr);
static int is_name(rtx_ptr expr);
static int match_argument(
    int argConstraint, int argValue, const rtx_ptr arg
);
static void compose_argument(char text[], rtx_ptr arg, int argType);
static void map_reg_to_hreg(int pseudoReg, int hreg);
static int try_to_free(int hreg, int line_number);
static int remap_pseudoreg(int pseudoreg, int new_hreg);
static int reg_in_stack_slot(int pseudoreg);
static void dump_register_allocation();
static int find_pseudoreg(int hreg, int line_number);

/* Reload an argument, if instruction needs it */
static int reload_arg(
    rtx_ptr arg, int argNumber,
    instr_def_cptr instr, 
    int *free_reg_set, int *reload_reg_set,
    int *reload_hreg, int *saved_hreg,
    int *reload_preg
);
/* Bits in a values returned by reload_arg */
#define ARG_RELOAD_NONE           0     /* Reload is not needed */
#define ARG_RELOAD_HREG           1     /* Arg. is reload in hreg */
#define ARG_RELOAD_SAVED_IN_STACK 2     /* Reload register saved in stack */
#define ARG_RELOAD_SAVED_IN_HREG  4     /* Reload register saved in register */

void genasm_init() {
    /* Open output file */
    if (output_file_name[0] == 0)
        strcpy(output_file_name, "out.s");

    output_file = fopen(output_file_name, "w");
    if (output_file == NULL) {
        perror("Could not open the output file"); exit(1);
    }

    /* Preambula */
    genasm_write_file_preambula();
}

void genasm_terminate() {
    genasm_write_file_postambula();
    fclose(output_file); output_file = NULL;
}

void genasm_write_file_preambula() {
    genasm_write_text(
        file_preambula,
        ARG_ESCAPE_CHARACTER,
        input_file_name, NULL, NULL
    );
}

void genasm_write_file_postambula() {
    genasm_write_text(
        file_postambula,
        ARG_ESCAPE_CHARACTER,
        NULL, NULL, NULL
    );
}

static void scan_register_usage_rtx(rtx_ptr expr, int line_number, int level) {
    rtx_ptr l, r;
    if (expr == NULL)
        return;
    if (level > 0 && expr->oper == RTX_REG) {
        int reg = expr->reg;
        assert(reg <= gen_lastregister());
        if (pseudo_reg_first_usage[reg] < 0)
            pseudo_reg_first_usage[reg] = line_number;
        pseudo_reg_last_usage[reg] = line_number;
        return;
    }
    l = expr->left;
    r = expr->right;
    if (level == 0) {
        if (l != NULL) {
            if (l->oper == RTX_REG) {
                assert(l->reg <= gen_lastregister());
                if (pseudo_reg_assignment[l->reg] < 0) {
                    pseudo_reg_assignment[l->reg] = line_number;
                }
            } else if (l->left != NULL) {
                /* mem[reg[n]] = ... */
                scan_register_usage_rtx(l->left, line_number, level + 1);
            }
        }
    } else {
        scan_register_usage_rtx(l, line_number, level + 1);
    }
    scan_register_usage_rtx(r, line_number, level + 1);
}

static int scan_rtl_line(block_ptr line, void *line_number, long arg1) {
    int *ln = (int *) line_number;
    rtl_ptr l = (rtl_ptr) line;
    if (l->typ == RTL_INSTR) {
        scan_register_usage_rtx(l->expr, *ln, 0);
    }
    (*ln)++;
    return 0;
}

void scan_register_usage() {
    int line_number = 0;
    int lastreg = gen_lastregister();
    if (pseudo_reg_array_size < lastreg + 1) {
        pseudo_reg_array_size = lastreg + 1;
        pseudo_reg_assignment = (int *) realloc(
            pseudo_reg_assignment, pseudo_reg_array_size * sizeof(int)
        );
        pseudo_reg_first_usage = (int *) realloc(
            pseudo_reg_first_usage, pseudo_reg_array_size * sizeof(int)
        );
        pseudo_reg_last_usage = (int *) realloc(
            pseudo_reg_last_usage, pseudo_reg_array_size * sizeof(int)
        );
        pseudo_register_map = (int *) realloc(
            pseudo_register_map, pseudo_reg_array_size * sizeof(int)
        );
        pseudo_register_stack_slot = (int *) realloc(
            pseudo_register_stack_slot, pseudo_reg_array_size * sizeof(int)
        );
    }

    memset(pseudo_reg_assignment, -1, sizeof(int)*pseudo_reg_array_size);
    memset(pseudo_reg_first_usage, -1, sizeof(int)*pseudo_reg_array_size);
    memset(pseudo_reg_last_usage, -1, sizeof(int)*pseudo_reg_array_size);
    memset(pseudo_register_map, -1, sizeof(int)*pseudo_reg_array_size);
    memset(pseudo_register_stack_slot, -1, sizeof(int)*pseudo_reg_array_size);

    lst_loop(&rtl_function, &scan_rtl_line, &line_number, 0);
}

void genasm_generate_function() {
    int line_number = 0;
    stack_32_slots = 0;         /* Set of first 32 allocated stack slots */
    stack_additional_space = 0; /* In words, including first 32 slots */
    stack_max_additional_space = 0;

    stack_save_hardregs_depth = 0;
    stack_save_hardregs_maxdepth = 0;

    scan_register_usage();

    /* First pass: register allocation */
    /* genasm_allocate_registers(); */
    genasm_pass = 0;
    allocated_registers = 0;
    clobbered_registers = 0;
    line_number = 0;
    lst_loop(&rtl_function, &generate_rtl_line, &line_number, 0);
    dump_register_allocation();

    genasm_pass++;
    genasm_write_function_beg();

    /* Second pass: Assembler generation */
    memset(pseudo_register_map, -1, sizeof(int)*pseudo_reg_array_size);
    memset(pseudo_register_stack_slot, -1, sizeof(int)*pseudo_reg_array_size);
    stack_32_slots = 0;
    stack_additional_space = 0;
    stack_save_hardregs_depth = 0;
    allocated_registers = 0;
    line_number = 0;
    lst_loop(&rtl_function, &generate_rtl_line, &line_number, 0);

    genasm_write_function_end();
}

static int generate_rtl_line(block_ptr line, void *line_number, long arg1) {
    int *ln = (int *) line_number;
    rtl_ptr l = (rtl_ptr) line;
    detect_basic_block(l, *ln);
    if (l->typ == RTL_INSTR) {
        generate_rtl_instr(l, *ln);
    } else if (l->typ == RTL_GOTO) {
        generate_rtl_goto(l);
    } else if (l->typ == RTL_LABEL) {
        generate_rtl_label(l);
    }
    (*ln)++;
    return 0;
}

static void generate_rtl_goto(rtl_ptr line) {
    if (genasm_pass == 0) return;
    if (line->int_value2 == 0)
        genasm_write_goto(line->int_value);
    else
        genasm_write_branch(line->int_value, line->int_value2);
}

void genasm_write_goto(int label) {
    char lab[32];
    if (genasm_pass == 0) return;
    memset(lab, 0, 32);
    sprintf(lab, "%s%d", LABEL_PREFIX, label);
    genasm_write_line(
        goto_instruction,
        ARG_ESCAPE_CHARACTER,
        lab, NULL, NULL
    );
}

static void generate_rtl_label(rtl_ptr line) {
    char lab[32];
    if (genasm_pass == 0) return;
    memset(lab, 0, 32);
    sprintf(lab, "%s%d", LABEL_PREFIX, line->int_value);
    genasm_write_line(
        label_line,
        ARG_ESCAPE_CHARACTER,
        lab, NULL, NULL
    );
}

void genasm_write_branch(int label, int condition) {
    char lab[32];
    if (genasm_pass == 0) return;
    memset(lab, 0, 32);
    sprintf(lab, "%s%d", LABEL_PREFIX, label);
    genasm_write_line(
        branch_instruction,
        ARG_ESCAPE_CHARACTER,
        lab, genasm_cond_name(condition), NULL
    );
}

const char *genasm_cond_name(int condition) {
    int n = 0;
    if (condition == EQ)      n = 0;
    else if (condition == NE) n = 1;
    else if (condition == GT) n = 2;
    else if (condition == GE) n = 3;
    else if (condition == LT) n = 4;
    else if (condition == LE) n = 5;
    else if (condition == GTU) n = 6;
    else if (condition == GEU) n = 7;
    else if (condition == LTU) n = 8;
    else if (condition == LEU) n = 9;
    return asm_condition_names[n];
}

/* Work with register sets */
int regset_add(int set, int reg) {
    assert (0 <= reg && reg < NUM_REGISTERS_ALLOCATE);
    return (set | (1 << reg));
}

int regset_remove(int set, int reg) {
    return (set & ~(1 << reg));
}

int regset_contains(int set, int reg) {
    assert (0 <= reg && reg < NUM_REGISTERS_ALLOCATE);
    return ((set & (1 << reg)) != 0);
}

int regset_erase(int set) {
    return 0;
}

int regset_subtract(int set, int set1) {
    return (set & ~set1);
}

int regset_union(int set, int set1) {
    return (set | set1);
}

int regset_intersect(int set, int set1) {
    return (set & set1);
}

int regset_loop(
    int set,
    int (*action)(int reg, void *arg1, long arg2),
    void *arg1, long arg2
) {
    int reg;
    for (reg = 0; reg < NUM_REGISTERS_ALLOCATE; reg++) {
        int res = action(reg, arg1, arg2);
        if (res != 0)
            return res;
    }
    return 0;
}

int regset_reverse_loop(
    int set,
    int (*action)(int reg, void *arg1, long arg2),
    void *arg1, long arg2
) {
    int reg;
    for (reg = NUM_REGISTERS_ALLOCATE - 1; reg >= 0; reg++) {
        int res = action(reg, arg1, arg2);
        if (res != 0)
            return res;
    }
    return 0;
}

int regset_complement(int set) {
    return (set ^ REGISTER_FULL_SET);
}

int register_index(int reg_bit) {
    int i;
    for (i = 0; i < NUM_REGISTERS_ALLOCATE; i++) {
        if ((reg_bit & 1) != 0)
            return i;
        reg_bit >>= 1;
    }
    assert(0);
    return 0;
}

int register_bit(int reg_idx) {
    assert (0 <= reg_idx && reg_idx < NUM_REGISTERS_ALLOCATE);
    return (1 << reg_idx);
}

int regset_get_first(int set) {
    int i, bit;
    for (i = 0, bit = 1; i < NUM_REGISTERS_ALLOCATE; i++, bit <<= 1) {
        if ((bit & set) != 0)
            return i;
    }
    return (-1);
}

int regset_get_next(int set, int reg) {
    int i, bit;
    assert(0 <= reg && reg < NUM_REGISTERS_ALLOCATE);
    i = reg + 1;
    bit = (register_bit(reg) << 1);
    while (i < NUM_REGISTERS_ALLOCATE) {
        if ((bit & set) != 0)
            return i;
        i++; bit <<= 1;
    }
    return (-1);
}

int regset_get_last(int set) {
    int i, bit;
    for (
        i = NUM_REGISTERS_ALLOCATE-1, bit = (1 << (NUM_REGISTERS_ALLOCATE-1));
        i >= 0;
        i--, bit >>= 1
    ) {
        if ((bit & set) != 0)
            return i;
    }
    return (-1);
}

int regset_get_prev(int set, int reg) {
    int i, bit;
    assert(0 <= reg && reg < NUM_REGISTERS_ALLOCATE);
    i = reg - 1;
    bit = (register_bit(reg) >> 1);
    while (i >= 0) {
        if ((bit & set) != 0)
            return i;
        i--; bit >>= 1;
    }
    return (-1);
}

/* Stupid register allocation */
static void allocate_reg(int pseudoreg) {
    /* Look up free register */
    int free_regs = regset_complement(allocated_registers);
    if (free_regs != 0) {
        int i;
        for (i = 0; i < NUM_REGISTERS_ALLOCATE; i++) {
            int reg = REGISTER_ALLOCATION_ORDER[i];
            if (regset_contains(free_regs, reg)) {
                /* Success, we have found a free register */
                pseudo_register_map[pseudoreg] = reg;
                allocated_registers = regset_add(allocated_registers, reg);
                clobbered_registers = regset_add(clobbered_registers, reg);
                return;
            }
        }
        assert(0); /* Should not come here */
    } else {
        /* No free registers, allocate a stack slot */
        pseudo_register_map[pseudoreg] = PSEUDO_REGISTER_IN_STACK_SLOT;
        allocate_stack_slot(pseudoreg);
    }
}

void release_pseudoreg(int pseudoreg) {
    int hreg = pseudo_register_map[pseudoreg];
    if (hreg == PSEUDO_REGISTER_UNMAPPED) {
        /* assert(false); */
        return;
    } else if (hreg == PSEUDO_REGISTER_IN_STACK_SLOT) {
        int slot = pseudo_register_stack_slot[pseudoreg];
        assert(slot >= 0);
        release_stack_slot(slot);
    } else {
        assert(0 <= hreg && hreg < NUM_REGISTERS_ALLOCATE);
        allocated_registers = regset_remove(allocated_registers, hreg);
    }
}

static void release_regs_rtx(rtx_ptr expr, int line_number) {
    if (expr == NULL)
        return;
    if (expr->oper == RTX_REG) {
        int pseudoreg = expr->reg;
        if (pseudo_reg_last_usage[pseudoreg] <= line_number) {
            release_pseudoreg(pseudoreg);
        }
    } else if (expr->oper == RTX_HARDREG) {
        if (expr->hard_reg < NUM_REGISTERS_ALLOCATE) {
            allocated_registers = regset_remove(
                allocated_registers, expr->hard_reg
            );
        }
    } else {
        if (expr->left != NULL)
            release_regs_rtx(expr->left, line_number);
        if (expr->right != NULL)
            release_regs_rtx(expr->right, line_number);
    }
}

static void allocate_regs_rtx(rtx_ptr expr, int line_number) {
    if (expr->oper != RTX_SET)  /* Call instr, etc. */
        return;

    /* Release register that are not used AFTER this instruction */
    if (expr->right != NULL)
        release_regs_rtx(expr->right, line_number);

    if (expr->left != NULL && expr->left->oper == RTX_REG)
        allocate_reg(expr->left->reg);
}

static int allocate_regs_rtl_line(
    block_ptr line, void *line_num, long arg1
) {
    int *ln = (int *) line_num;
    int line_number = *ln;

    rtl_ptr l = (rtl_ptr) line;

    if (l->typ == RTL_INSTR) {
        allocate_regs_rtx(l->expr, line_number);
    }

    (*ln)++;
    return 0;
}

void genasm_allocate_registers() { /* First pass: register allocation */
    int line_number = 0;
    allocated_registers = 0;    /* Erase the set of allocated registers */
    clobbered_registers = 0;
    lst_loop(&rtl_function, &allocate_regs_rtl_line, &line_number, 0);
    dump_register_allocation();
}

static void dump_register_allocation() {
    if (compiler_flags & DUMP_RTL) {
        int reg, hreg;
        /* Dump pseudoregister map */
        int lastreg = gen_lastregister();

        fprintf(dump_file, "Pseudoregister map\n");
        for (reg = MIN_PSEUDOREG; reg <= lastreg; reg++) {
            hreg = pseudo_register_map[reg];
            if (hreg == PSEUDO_REGISTER_IN_STACK_SLOT) {
                fprintf(
                    dump_file, "  Pseudoreg %d in stack slot %d\n",
                    reg, pseudo_register_stack_slot[reg]
                );
            } else if (0 <= hreg && hreg < NUM_REGISTERS_ALLOCATE) {
                fprintf(
                    dump_file, "  Pseudoreg %d mapped to hreg %s\n",
                    reg, REGISTER_NAME[hreg]
                );
            } else {
                fprintf(
                    dump_file, "  Pseudoreg %d is not mapped\n", reg
                );
            }
        }
    }
}

static int reg_in_stack_slot(int pseudoreg) {
    return (pseudo_register_map[pseudoreg] == PSEUDO_REGISTER_IN_STACK_SLOT);
}

void allocate_stack_slot(int pseudoreg) {
    int slot = (-1);
    int i, bit;

    /* Try to find out the stack slot among the first 32 */
    for (i = 0, bit = 1; i < 32; i++, bit <<= 1) {
        if ((stack_32_slots & bit) == 0) {
            slot = i;
            stack_32_slots |= bit;
            if (stack_additional_space < slot + 1)
                stack_additional_space = slot + 1;
            if (stack_additional_space > stack_max_additional_space)
                stack_max_additional_space = stack_additional_space;
            break;
        }
    }

    if (slot < 0) {
        /* All 32 lower slots are allocated */
        slot = stack_additional_space;
        stack_additional_space++;
        if (stack_additional_space > stack_max_additional_space)
            stack_max_additional_space = stack_additional_space;
    }
    pseudo_register_stack_slot[pseudoreg] = slot;
}

void release_stack_slot(int slot) {
    assert(0 <= slot && slot < stack_additional_space);
    if (slot < 32) {
        stack_32_slots &= ~(1 << slot);
    }
    if (stack_additional_space == slot + 1)
        stack_additional_space--;
}

void genasm_write_text(
    const char * const lines[], /* Sequence of lines to write */
    int escape_character,       /* Character marking a place to insert */
    const char *insert0,        /* First pattern to insert */
    const char *insert1,        /* Second pattern to insert */
    const char *insert2         /* Third pattern to insert */
) {
    int i = 0;
    while (lines[i] != NULL) {
        genasm_write_line(
            lines[i], escape_character,
            insert0, insert1, insert2
        );
        i++;
    }
}

void genasm_write_line(
    const char *str,            /* Line to write */
    int escape_character,       /* Character marking a place to insert */
    const char *insert0,        /* First pattern to insert */
    const char *insert1,        /* Second pattern to insert */
    const char *insert2         /* Third pattern to insert */
) {
    char line[256];
    if (str == NULL) return;
    strncpy(line, str, 224); line[224] = 0;

    /* Replace all templates in a line */
    genasm_substitute(
        line, escape_character, insert0, insert1, insert2
    );

    fputs(line, output_file);
}

void genasm_substitute(
    char *line,
    int escape_character,       /* Character marking a place to insert */
    const char *insert0,        /* First pattern to insert */
    const char *insert1,        /* Second pattern to insert */
    const char *insert2         /* Third pattern to insert */
) {
    /* Replace all templates in a line */
    char *templ;
    int l, lins, ltail;
    l = strlen(line);

    templ = line;
    while (
        templ < line + l - 1 &&
        (templ = strchr(templ, escape_character)) != NULL
    ) {
        const char *ins = insert0;
        if (templ[1] == '1') ins = insert1;
        else if (templ[1] == '2') ins = insert2;
        assert(ins != NULL);
        lins = strlen(ins);
        ltail = l - ((templ + 2) - line);
        if (ltail > 0)
            memmove(templ + lins, templ + 2, ltail);
        memmove(templ, ins, lins);
        l += lins - 2;
        line[l] = 0;
        templ += lins;
    }
}

void genasm_write_function_beg() {
    char number[32];
    name_ptr n;
    int reg;
    int frameSize;

    genasm_write_text(
        function_preambula,
        ARG_ESCAPE_CHARACTER,
        current_function->name, NULL, NULL
    );

    /* Calculate size of local variables */
    n = (name_ptr) local_table.next;
    frameSize = 0;
    while (n != (name_ptr) &local_table) {
        if (!(n->typ.decl & DECL_FUN)) {
            frameSize += n->typ.size;
            if (frameSize % 4 != 0)
                frameSize += (4 - frameSize % 4); /* Align to word boundary */
        }
        n = (name_ptr) n->link.next;
    }

    stack_slots_offset = (-4) - frameSize;
    frameSize += 4 * stack_max_additional_space;
    frameSize += 4 * stack_save_hardregs_maxdepth;

    if (frameSize > 0) {
        memset(number, 0, 32);
        sprintf(number, "%d", frameSize);
        genasm_write_line(
            stack_advance_instruction,
            ARG_ESCAPE_CHARACTER,
            number, NULL, NULL
        );
    }

    /* Save clobbered registers */
    reg = regset_get_first(clobbered_registers);
    while (reg >= 0) {
        if (regset_contains(SAVED_REGISTERS, reg)) {
            generate_push_hreg(reg);
        }
        reg = regset_get_next(clobbered_registers, reg);
    }
}

static int function_close_label = 0;

void genasm_write_function_end() {
    char close_label[32];
    int reg;

    /* Restore registers */
    reg = regset_get_last(clobbered_registers);
    while (reg >= 0) {
        if (regset_contains(SAVED_REGISTERS, reg)) {
            generate_pop_hreg(reg);
        }
        reg = regset_get_prev(clobbered_registers, reg);
    }

    function_close_label++;
    memset(close_label, 0, 32);
    sprintf(close_label, "%s%d",
        FUNCTION_CLOSE_LABEL_PREFIX, function_close_label);

    genasm_write_text(
        function_postambula,
        ARG_ESCAPE_CHARACTER,
        close_label, current_function->name, NULL
    );
}

static void generate_rtl_instr(rtl_ptr line, int line_number) {
    rtx_ptr e = line->expr;
    rtx_ptr result = e->left;
    rtx_ptr action = e->right;  /* Action or arg of move command */
    int free_regs;
    rtx_ptr arg0, arg1;
    int res_reg, hreg;
    int regs_to_save;
    char res_text[64], *res_txt;
    char arg0_text[64], *arg0_txt;
    char arg1_text[64], *arg1_txt;
    instr_def_cptr instDef;

    /* For register reload */
    int reload_action[3];
    int reload_preg[3];
    int reload_hreg[3];
    int saving_hreg[3];
    int reload_saved_preg_map[3];
    int free_regs_for_reload;
    int reload_regs;
    rtx arg_hreg;
    rtx_ptr res_ptr;
    int i;

    assert(result != NULL && action != NULL);

    /* Define arguments of instruction */
    if (is_reg_hreg(action) || is_mem(action) || is_bytemem(action) ||
        is_const(action) || is_name(action)) {
        arg0 = action;
        arg1 = NULL;
    } else {
        arg0 = action->left;
        arg1 = action->right;
    }

    free_regs = regset_complement(allocated_registers);

    instDef = match_instr(e);
    assert(instDef != NULL);

    /* Allocate a register for result */
    if (is_reg(result)) { /* Result is a pseudo register */
        int res_allocated = 0;
        res_reg = result->reg;
        if (pseudo_reg_assignment[res_reg] < line_number) {
            /* Hardreg for this pseudoreg was allocated before */
            res_allocated = 1;
        } else if (instDef->res_set != 0) {
            hreg = regset_get_first(instDef->res_set);
            if (regset_contains(free_regs, hreg)) {
                map_reg_to_hreg(res_reg, hreg);
            } else {
                if (try_to_free(hreg, line_number))
                    map_reg_to_hreg(res_reg, hreg);
                else
                    allocate_reg(res_reg); /* Stupid allocation */
            }
            res_allocated = 1;
        } else if (instDef->res_in_arg == RES_IN_ARG0 ||
            instDef->res_in_arg == RES_IN_ARG1) {
            /* Result is in argument */
            int hreg;
            rtx_ptr arg = arg0;
            if (instDef->res_in_arg == RES_IN_ARG1)
                arg = arg1;
            assert(is_reg_hreg(arg));
            if (is_reg(arg)) {
                hreg = pseudo_register_map[arg->reg];
            } else {
                assert(is_hreg(arg));
                hreg = arg->hard_reg;
            }
            if (regset_contains(free_regs, hreg) ||
                (is_reg(arg) &&
                pseudo_reg_last_usage[arg->reg] <= line_number) ||
                is_hreg(arg)
            ) {
                map_reg_to_hreg(res_reg, hreg);
            } else {
                if (try_to_free(hreg, line_number))
                    map_reg_to_hreg(res_reg, hreg);
                else
                    allocate_reg(res_reg); /* Stupid allocation */
            }
            res_allocated = 1;
        }
        if (!res_allocated) {
            int res_regs = (free_regs & ~(instDef->scratch_regs));
            if (res_regs != 0) {
                hreg = regset_get_first(res_regs);
                map_reg_to_hreg(res_reg, hreg);
            } else {
                allocate_reg(res_reg); /* Stupid allocation */
            }
        }
    } else if (is_hreg(result)) {
        hreg = result->hard_reg;
        if (hreg < NUM_REGISTERS_ALLOCATE && 
            !regset_contains(free_regs, hreg)) {

            /* Maybe, this register will be freed after this instruction? */
            /* If yes, we need not to free it */
            if (!(
                (is_reg(arg0) &&
                pseudo_register_map[arg0->reg] == hreg &&
                pseudo_reg_last_usage[arg0->reg] <= line_number)
                ||
                (is_reg(arg1) &&
                pseudo_register_map[arg1->reg] == hreg &&
                pseudo_reg_last_usage[arg1->reg] <= line_number)
            ))
                try_to_free(hreg, line_number);
        }
    }

    /* Save registers that can be destroyed by instruction */
    regs_to_save = instDef->scratch_regs;
    if (instDef->res_in_arg == RES_IN_ARG0) {
        /* Result is in first argument */
        assert(is_reg_hreg(arg0));
        if (is_reg(arg0)) {
            hreg = pseudo_register_map[arg0->reg];
        } else {
            hreg = arg0->hard_reg;
        }
        regs_to_save = regset_add(regs_to_save, hreg);
    } else if (instDef->res_in_arg == RES_IN_ARG1) {
        /* Result is in second argument */
        assert(is_reg_hreg(arg1));
        if (is_reg(arg1)) {
            hreg = pseudo_register_map[arg1->reg];
        } else {
            hreg = arg1->hard_reg;
        }
        regs_to_save = regset_add(regs_to_save, hreg);
    }
    if (is_hreg(result)) {
        hreg = result->hard_reg;
        if (hreg < NUM_REGISTERS_ALLOCATE &&
            regset_contains(allocated_registers, hreg)) {
            regs_to_save = regset_add(regs_to_save, hreg);
        }
    }
    /* Do not save result! */
    if (is_reg_hreg(result)) {
        if (is_reg(result))
            hreg = pseudo_register_map[result->reg];
        else
            hreg = result->hard_reg;
        if (hreg < NUM_REGISTERS_ALLOCATE) {
            regs_to_save = regset_remove(regs_to_save, hreg);
        }
    }
    regs_to_save = regset_intersect(regs_to_save, allocated_registers);
    free_regs = regset_complement(allocated_registers);
    free_regs = regset_subtract(free_regs, regs_to_save);

    hreg = regset_get_first(regs_to_save);
    while (hreg >= 0) {
        /* At first, try to re-map a pseudoreg that is mapped to hreg */
        int remapped = 0;
        if (free_regs != 0) {
            int newReg = regset_get_first(free_regs);
            while (!remapped && newReg >= 0) {
                int preg = find_pseudoreg(hreg, line_number);
                /*... assert(preg >= 0); ...*/
                if (preg >= 0 && remap_pseudoreg(preg, newReg)) {
                    free_regs = regset_remove(free_regs, newReg);
                    regs_to_save = regset_remove(regs_to_save, hreg);
                    remapped = 1;
                } else {
                    newReg = regset_get_next(free_regs, newReg);
                }
            }
        }
        if (!remapped) {
            /* If we could not remap, just push hreg in stack */
            generate_save_hreg(hreg);
        }
        hreg = regset_get_next(regs_to_save, hreg);
    }
    /* End of saving registers */

    /* Reload registers, if necessary */
    reload_action[0] = reload_action[1] = reload_action[2] = ARG_RELOAD_NONE;
    free_regs_for_reload = regset_complement(allocated_registers);
    reload_regs = 0;
    memset(&arg_hreg, 0, sizeof(arg_hreg));
    arg_hreg.oper = RTX_HARDREG;

    assert(arg0 != NULL);
    /* Reload, if necessary */
    reload_action[0] = reload_arg(
        arg0, 0,
        instDef,
        &free_regs_for_reload, &reload_regs,
        &(reload_hreg[0]), &(saving_hreg[0]),
        &(reload_preg[0])
    );
    /*... reload_preg[0] = arg0->reg; ...*/
    if (reload_action[0] != ARG_RELOAD_NONE) {
        /*...
        arg_hreg.hard_reg = reload_hreg[0];
        arg0 = &arg_hreg;
        ...*/
        reload_saved_preg_map[0] = pseudo_register_map[reload_preg[0]];
        pseudo_register_map[reload_preg[0]] = reload_hreg[0];
    }

    /* Consider separately "move reg to byte memory" instruction */
    if (instDef == INSTR_MOV_REG_BYTE) {
        assert(is_reg(arg0) || is_hreg(arg0) || is_const(arg0));
        arg0->modifiers |= RTX_MODBYTE;
    }

    compose_argument(arg0_text, arg0, instDef->arg0_constraint);
    arg0_txt = arg0_text;

    if (arg1 != NULL) {
        /* Reload, if necessary */
        reload_action[1] = reload_arg(
            arg1, 1,
            instDef,
            &free_regs_for_reload, &reload_regs,
            &(reload_hreg[1]), &(saving_hreg[1]),
            &(reload_preg[1])
        );
        /*... reload_preg[1] = arg1->reg; ...*/
        if (reload_action[1] != ARG_RELOAD_NONE) {
            /*...
            memset(&arg_hreg, 0, sizeof(arg_hreg));
            arg_hreg.oper = RTX_HARDREG;
            arg_hreg.hard_reg = reload_hreg[1];
            arg1 = &arg_hreg;
            ...*/
            reload_saved_preg_map[1] = pseudo_register_map[reload_preg[1]];
            pseudo_register_map[reload_preg[1]] = reload_hreg[1];
        }
        compose_argument(arg1_text, arg1, instDef->arg1_constraint);
        arg1_txt = arg1_text;
    } else {
        arg1_txt = NULL;
    }

    res_ptr = result;
    /* Reload, if necessary */
    reload_action[2] = reload_arg(
        res_ptr, 2,
        instDef,
        &free_regs_for_reload, &reload_regs,
        &(reload_hreg[2]), &(saving_hreg[2]),
        &(reload_preg[2])
    );
    /*... reload_preg[2] = res_ptr->reg; ...*/
    if (reload_action[2] != ARG_RELOAD_NONE) {
        /*...
        memset(&arg_hreg, 0, sizeof(arg_hreg));
        arg_hreg.oper = RTX_HARDREG;
        arg_hreg.hard_reg = reload_hreg[2];
        res_ptr = &arg_hreg;
        ...*/
        reload_saved_preg_map[2] = pseudo_register_map[reload_preg[2]];
        pseudo_register_map[reload_preg[2]] = reload_hreg[2];
    }
    compose_argument(res_text, res_ptr, instDef->res_constraint);
    res_txt = res_text;

    if (genasm_pass > 0) {
        /* Do not generate a "move to itself" instruction */
        if ((instDef != INSTR_MOV_REG_ANY &&
            instDef != INSTR_MOV_ANY_REG) ||
            strcmp(arg0_txt, res_txt) != 0
        ) {
            /* Generate an instruction */
            genasm_write_line(
                instDef->templ,
                ARG_ESCAPE_CHARACTER,
                arg0_txt, arg1_txt, res_txt
            );
        }
    }

    /* In case of 2-address instruction, copy result to
       its destination */
    if (((instDef->res_in_arg == RES_IN_ARG0 && is_reg_hreg(arg0)) ||
        (instDef->res_in_arg == RES_IN_ARG1 && is_reg_hreg(arg1))) &&
        is_reg(result)) {
        int preg = result->reg;
        int hreg = pseudo_register_map[preg];
        int arghreg;

        /*...
        if (reload_action[2] & ARG_RELOAD_HREG) {
            hreg = reload_hreg[2];
        }
        ...*/
        if (instDef->res_in_arg == RES_IN_ARG0) {
            if (is_hreg(arg0)) {
                arghreg = arg0->hard_reg;
            /*...
            } else if (reload_action[0] & ARG_RELOAD_HREG) {
                arghreg = reload_hreg[0];
            ...*/
            } else {
                arghreg = pseudo_register_map[arg0->reg];
            }
        } else if (instDef->res_in_arg == RES_IN_ARG1) {
            if (is_hreg(arg1)) {
                arghreg = arg1->hard_reg;
            /*...
            } else if (reload_action[1] & ARG_RELOAD_HREG) {
                arghreg = reload_hreg[1];
            ...*/
            } else {
                arghreg = pseudo_register_map[arg1->reg];
            }
        }
        generate_move_hregs(arghreg, hreg);
    }

    /* In case of reloading, copy result to its pseudo register */
    if ((reload_action[2] & ARG_RELOAD_HREG) && is_reg(result)) {
        pseudo_register_map[reload_preg[2]] = reload_saved_preg_map[2];
        generate_move_hreg_to_pseudoreg(
            reload_hreg[2], reload_preg[2]
        );
    }

    /* Restore reload registers */
    for (i = 2; i >= 0; i--) {
        if (reload_action[i] == ARG_RELOAD_NONE) continue;
        pseudo_register_map[reload_preg[i]] = reload_saved_preg_map[i];
        if (reload_action[i] & ARG_RELOAD_SAVED_IN_HREG) {
            generate_move_hregs(
                saving_hreg[i], reload_hreg[i]
            );
        } else if (reload_action[i] & ARG_RELOAD_SAVED_IN_STACK) {
            generate_restore_hreg(reload_hreg[i]);
        }
    }

    /* Restore saved registers */
    hreg = regset_get_last(regs_to_save);
    while (hreg >= 0) {
        generate_restore_hreg(hreg);
        hreg = regset_get_prev(regs_to_save, hreg);
    }
    /* End of saving registers */

    release_regs_rtx(action, line_number); /* Release unused registers */

    /* If result is a hard register, mark it as allocated */
    if (is_hreg(result) && result->hard_reg < NUM_REGISTERS_ALLOCATE) {
        allocated_registers = regset_add(
            allocated_registers, result->hard_reg
        );
    }
}

static int try_to_free(int hreg, int line_number) {
    int newReg;
    int freeRegs;
    int pseudoReg;
    if (!regset_contains(allocated_registers, hreg))
        return 1;
    pseudoReg = find_pseudoreg(hreg, line_number);
    if (pseudoReg < 0) {
        /* No pseudoregisters use this hard register. */
        /* Maybe, it is allocated for result of instruction, etc */
        return 1;       /* Return OK */
    }

    /* If a scope of a pseudoregister goes outside of basic block,
       we cannot re-map it */
    if (pseudo_reg_last_usage[pseudoReg] >= basic_block_end)
        return 0;

    /* Find out a free register */
    freeRegs = regset_complement(allocated_registers);
    freeRegs = regset_remove(freeRegs, hreg);
    if (freeRegs == 0)
        return 0;
    newReg = regset_get_first(freeRegs);
    return remap_pseudoreg(pseudoReg, newReg);
}

static int remap_pseudoreg(int pseudoreg, int new_hreg) {
    int hreg = pseudo_register_map[pseudoreg];
    assert(hreg >= 0 && hreg < NUM_REGISTERS_ALLOCATE);
    assert(new_hreg >= 0 && new_hreg < NUM_REGISTERS_ALLOCATE);
    if (hreg == new_hreg)
        return 1; /* Nothing to do */

    /* The pseudoregister scope must be in the current basic block */
    if (pseudo_reg_last_usage[pseudoreg] >= basic_block_end)
        return 0;

    pseudo_register_map[pseudoreg] = new_hreg;
    allocated_registers = regset_remove(allocated_registers, hreg);
    allocated_registers = regset_add(allocated_registers, new_hreg);
    clobbered_registers = regset_add(clobbered_registers, new_hreg);
    if (genasm_pass > 0)
        generate_move_hregs(hreg, new_hreg);
    return 1;
}

static void map_reg_to_hreg(int pseudoReg, int hreg) {
    pseudo_register_map[pseudoReg] = hreg;
    allocated_registers = regset_add(allocated_registers, hreg);
    clobbered_registers = regset_add(clobbered_registers, hreg);
}

static void generate_save_hreg(int hreg) {
    stack_save_hardregs_depth++;
    if (stack_save_hardregs_depth > stack_save_hardregs_maxdepth)
        stack_save_hardregs_maxdepth = stack_save_hardregs_depth;

    if (genasm_pass > 0) {
        char number[64];
        char restext[64];
        int offset = stack_slots_offset -
            stack_max_additional_space * 4 + 4 -
            stack_save_hardregs_depth * 4;
        memset(number, 0, 64);
        memset(restext, 0, 64);

        /* Compose memory argument of "mov" instruction */
        sprintf(number, "%d", offset);
        strcpy(restext, arg_mem_reg_const);
        genasm_substitute(
            restext,
            ARG_ESCAPE_CHARACTER,
            REGISTER_NAME[FP], number, NULL
        );

        /* Compose "mov" instruction */
        genasm_write_line(
            INSTR_MOV_REG_ANY->templ,
            ARG_ESCAPE_CHARACTER,
            REGISTER_NAME[hreg], NULL, restext
        );
    }
}

static void generate_restore_hreg(int hreg) {
    int depth = stack_save_hardregs_depth;
    assert(depth > 0);
    stack_save_hardregs_depth--;

    if (genasm_pass > 0) {
        char number[64];
        char argtext[64];
        int offset = stack_slots_offset -
            stack_max_additional_space * 4 + 4 -
            depth * 4;
        memset(number, 0, 64);
        memset(argtext, 0, 64);

        /* Compose memory argument of "mov" instruction */
        sprintf(number, "%d", offset);
        strcpy(argtext, arg_mem_reg_const);
        genasm_substitute(
            argtext,
            ARG_ESCAPE_CHARACTER,
            REGISTER_NAME[FP], number, NULL
        );

        /* Compose "mov" instruction */
        genasm_write_line(
            INSTR_MOV_ANY_REG->templ,
            ARG_ESCAPE_CHARACTER,
            argtext, NULL, REGISTER_NAME[hreg]
        );
    }
}

static void generate_push_hreg(int hreg) {
    if (genasm_pass == 0) return;
    genasm_write_line(
        INSTR_PUSH->templ,
        ARG_ESCAPE_CHARACTER,
        REGISTER_NAME[hreg], NULL, NULL
    );
}

static void generate_pop_hreg(int hreg) {
    if (genasm_pass == 0) return;
    genasm_write_line(
        INSTR_POP->templ,
        ARG_ESCAPE_CHARACTER,
        REGISTER_NAME[hreg], NULL, NULL
    );
}

static void generate_move_hregs(int hreg_src, int hreg_dst) {
    if (genasm_pass == 0) return;
    if (hreg_src == hreg_dst)
        return;
    genasm_write_line(
        INSTR_MOV_REG_ANY->templ,
        ARG_ESCAPE_CHARACTER,
        REGISTER_NAME[hreg_src], NULL, REGISTER_NAME[hreg_dst]
    );
}

static void generate_move_pseudoreg_to_hreg(int pseudoreg, int hreg) {
    int srcReg = pseudo_register_map[pseudoreg];
    if (genasm_pass == 0) return;
    if (srcReg != PSEUDO_REGISTER_IN_STACK_SLOT) {
        generate_move_hregs(srcReg, hreg);
    } else {
        char argText[64];
        rtx arg;
        arg.oper = RTX_REG; arg.reg = pseudoreg;
        compose_argument(argText, &arg, ARG_REGISTER);
        genasm_write_line(
            INSTR_MOV_ANY_REG->templ,
            ARG_ESCAPE_CHARACTER,
            argText, NULL, REGISTER_NAME[hreg]
        );
    }
}

static void generate_move_hreg_to_pseudoreg(int hreg, int pseudoreg) {
    int dstReg = pseudo_register_map[pseudoreg];
    if (genasm_pass == 0) return;
    if (dstReg != PSEUDO_REGISTER_IN_STACK_SLOT) {
        generate_move_hregs(hreg, dstReg);
    } else {
        char argText[64];
        rtx arg;
        arg.oper = RTX_REG; arg.reg = pseudoreg;
        compose_argument(argText, &arg, ARG_REGISTER);
        genasm_write_line(
            INSTR_MOV_REG_ANY->templ,
            ARG_ESCAPE_CHARACTER,
            REGISTER_NAME[hreg], NULL, argText
        );
    }
}

static instr_def_cptr match_instr(rtx_ptr expr) {
    instr_def_cptr instDef;
    int oper = expr->oper;
    int numArgs;
    rtx_ptr result = expr->left;
    rtx_ptr action = expr->right;

    /* Consider separately "mov" and "push" instructions */
    if (oper == RTX_SET && is_reg_hreg(expr->left) &&
        (is_reg_mem(expr->right) || is_bytemem(expr->right) ||
        is_const(expr->right) || is_name(expr->right))) {
        /* Move to register instruction */
        if (expr->right->modifiers & RTX_MODBYTE) {
            return INSTR_MOV_BYTE_REG;
        } else {
            return INSTR_MOV_ANY_REG;
        }
    } else if (oper == RTX_SET && 
        (is_mem(expr->left) || is_bytemem(expr->left)) &&
        (is_reg_hreg(expr->right) || is_const(expr->right))) {
        /* Move to memory instruction */
        if (expr->left->modifiers & RTX_MODBYTE) {
            return INSTR_MOV_REG_BYTE;
        } else if (expr->left->left != NULL &&
            expr->left->left->oper == RTX_PREDEC &&
            expr->left->left->left->oper == RTX_HARDREG &&
            expr->left->left->left->hard_reg == SP) {
            return INSTR_PUSH;
        } else if (expr->left->modifiers & RTX_MODBYTE) {
            return INSTR_MOV_REG_BYTE;
        } else {
            return INSTR_MOV_REG_ANY;
        }
    }

    assert (expr->left != NULL && expr->right != NULL);
    numArgs = 1;        /* Result is always present */
    if (action->left != NULL) numArgs++;
    if (action->right != NULL) numArgs++;

    /* General case */
    for (instDef = instr_set; instDef->rtx_oper != RTX_NONE; instDef++) {
        if (instDef->rtx_oper != action->oper || 
            instDef->rtx_num_args != numArgs)
            continue;
        if (
            match_argument(
                instDef->rtx_res_constraint,
                instDef->rtx_res_value,
                result
            ) &&
            match_argument(
                instDef->rtx_arg0_constraint,
                instDef->rtx_arg0_value,
                action->left
            ) &&
            match_argument(
                instDef->rtx_arg1_constraint,
                instDef->rtx_arg1_value,
                action->right
            )
        ) {
            return instDef;
        }
    }
    assert(0);
    return NULL;
}

static int match_argument(
    int argConstraint, int argValue, const rtx_ptr expr
) {
    if (argConstraint == ARG_RTX_NONE && expr == NULL)
        return 1;
    switch (argConstraint) {
        case ARG_RTX_REGISTER:
            return is_reg_hreg(expr);
        case ARG_RTX_MEMORY:
            return is_mem(expr);
        case ARG_RTX_BYTE_MEMORY:
            return is_bytemem(expr);
        case ARG_RTX_ANY:
            return (is_reg_mem(expr) || is_const(expr) || is_name(expr));
        case ARG_RTX_HARDREG:
            return (expr->oper == RTX_HARDREG && expr->hard_reg == argValue);
        case ARG_RTX_CC0:
            return (expr->oper == RTX_HARDREG && expr->hard_reg == CC0);
        case ARG_RTX_ANY_CONST:
            return (expr->oper == RTX_ICON);
        case ARG_RTX_CONST:
            return (expr->oper == RTX_ICON && expr->int_value == argValue);
        case ARG_RTX_NAME:
            return (expr->oper == RTX_NAME);
        case ARG_RTX_MEMREG_PREDEC:
            /* Now used only for push */
            return (
                expr->oper == RTX_MEM &&
                expr->left->oper == RTX_PREDEC &&
                expr->left->left->oper == RTX_HARDREG &&
                expr->left->left->hard_reg == argValue
            );
        case ARG_RTX_MEMREG_POSTINC:
            /* Now used only for pop */
            return (
                expr->oper == RTX_MEM &&
                expr->left->oper == RTX_POSTINC &&
                expr->left->left->oper == RTX_HARDREG &&
                expr->left->left->hard_reg == argValue
            );
    }
    return 0;
}

static int is_reg_mem(rtx_ptr expr) {
    return (is_reg_hreg(expr) || is_mem(expr));
}

static int is_reg(rtx_ptr expr) {
    return (expr != NULL && expr->oper == RTX_REG);
}

static int is_hreg(rtx_ptr expr) {
    return (
        expr != NULL && expr->oper == RTX_HARDREG &&
        expr->hard_reg != CC0
    );
}

static int is_reg_hreg(rtx_ptr expr) {
    return (is_reg(expr) || is_hreg(expr));
}

static int is_mem(rtx_ptr expr) {
    return (
        expr != NULL && expr->oper == RTX_MEM && 
        !(expr->modifiers & RTX_MODBYTE)
    );
}

static int is_bytemem(rtx_ptr expr) {
    return (
        expr != NULL && expr->oper == RTX_MEM && 
        (expr->modifiers & RTX_MODBYTE)
    );
}

static int is_const(rtx_ptr expr) {
    return (expr != NULL && expr->oper == RTX_ICON);
}

static int is_name(rtx_ptr expr) {
    return (expr != NULL && expr->oper == RTX_NAME);
}

static void compose_argument(char text[], rtx_ptr arg, int argType) {
    int hreg;
    char number[64];
    memset(text, 0, 32);
    memset(number, 0, 64);
    if (arg->oper == RTX_REG) {
        hreg = pseudo_register_map[arg->reg];
        if (hreg == PSEUDO_REGISTER_IN_STACK_SLOT) {
            int offset = stack_slots_offset -
                pseudo_register_stack_slot[arg->reg] * 4;
            assert(pseudo_register_stack_slot[arg->reg] >= 0);
            sprintf(number, "%d", offset);
            strcpy(text, arg_mem_reg_const);
            genasm_substitute(
                text,
                ARG_ESCAPE_CHARACTER,
                REGISTER_NAME[FP], number, NULL
            );
        } else {
            assert(hreg >= 0 && hreg < NUM_REGISTERS_ALLOCATE);
            if (arg->modifiers & RTX_MODBYTE) {
                strcpy(text, BYTE_REGISTER_NAME[hreg]);
            } else {
                strcpy(text, REGISTER_NAME[hreg]);
            }
        }
    } else if (arg->oper == RTX_HARDREG && arg->hard_reg != CC0) {
        if (arg->modifiers & RTX_MODBYTE) {
            strcpy(text, BYTE_REGISTER_NAME[arg->hard_reg]);
        } else {
            strcpy(text, REGISTER_NAME[arg->hard_reg]);
        }
    } else if (arg->oper == RTX_NAME) {
        if (argType == ARG_PLAIN_NAME) {
            strcpy(text, arg->string_value);
        } else {
            if (arg->modifiers & RTX_MODGLOBAL)
                strcpy(text, arg_global_name);
            else
                strcpy(text, arg_name);
            genasm_substitute(
                text,
                ARG_ESCAPE_CHARACTER,
                arg->string_value, NULL, NULL
            );
        }
    } else if (arg->oper == RTX_ICON) {
        sprintf(number, "%d", arg->int_value);
        if (argType == ARG_PLAIN_CONST) {
            strcpy(text, number);
        } else {
            strcpy(text, arg_const);
            genasm_substitute(
                text,
                ARG_ESCAPE_CHARACTER,
                number, NULL, NULL
            );
        }
    } else if (arg->oper == RTX_MEM) {
        if (is_reg_hreg(arg->left)) {
            if (is_reg(arg->left))
                hreg = pseudo_register_map[arg->left->reg];
            else
                hreg = arg->left->hard_reg;
            strcpy(text, arg_mem_reg);
            genasm_substitute(
                text,
                ARG_ESCAPE_CHARACTER,
                REGISTER_NAME[hreg], NULL, NULL
            );
        } else if (arg->left->oper == RTX_NAME) {
            if (arg->left->modifiers & RTX_MODGLOBAL)
                strcpy(text, arg_mem_global_name);
            else
                strcpy(text, arg_mem_name);
            genasm_substitute(
                text,
                ARG_ESCAPE_CHARACTER,
                arg->left->string_value, NULL, NULL
            );
        } else if (arg->left->oper == RTX_PLUS) {
            /* reg + const */
            rtx_ptr r, c;
            r = arg->left->left;
            c = arg->left->right;
            if (is_reg_hreg(c)) {
                r = arg->left->right;
                c = arg->left->left;
            }
            if (is_reg(r))
                hreg = pseudo_register_map[r->reg];
            else
                hreg = r->hard_reg;
            sprintf(number, "%d", c->int_value);
            strcpy(text, arg_mem_reg_const);
            genasm_substitute(
                text,
                ARG_ESCAPE_CHARACTER,
                REGISTER_NAME[hreg], number, NULL
            );
        }
    }
}

static int find_pseudoreg(int hreg, int line_number) {
    int preg;
    int lastreg = gen_lastregister();
    for (preg = MIN_PSEUDOREG; preg <= lastreg; preg++) {
        int hreg = pseudo_register_map[preg];
        int pregAssign = pseudo_reg_assignment[preg];
        if (pregAssign >= line_number)
            break;
        if (0 <= hreg && hreg < NUM_REGISTERS_ALLOCATE && 
            pseudo_reg_last_usage[preg] >= line_number) {
            return preg;
        }
    }
    return (-1);
}

/* Bits in a values returned by reload_arg
#define ARG_RELOAD_NONE           0     * Reload is not needed *
#define ARG_RELOAD_HREG           1     * Arg. is reload in hreg *
#define ARG_RELOAD_SAVED_IN_STACK 2     * Reload register saved in stack *
#define ARG_RELOAD_SAVED_IN_HREG  4     * Reload register saved in register *
*/
static int reload_arg(
    rtx_ptr arg, int argNumber,
    instr_def_cptr instr, 
    int *free_reg_set, int *reload_reg_set,
    int *reload_hreg, int *saved_hreg,
    int *reload_preg
) {
    int argConstraint;
    int regSet;
    int ret = ARG_RELOAD_NONE;
    int possibleRegs;
    int bestRegs;
    int instrRegs = 0; /* All special registers used by other arguments */
    int preg, hreg;

    if (is_hreg(arg))
        return ret;

    if (argNumber == 0) {
        argConstraint = instr->arg0_constraint; /* First argument */
        regSet = instr->arg0_set;               /* Possible registers */
        if (instr->arg1_constraint == ARG_REGISTER_SET ||
            instr->arg1_constraint == ARG_BYTE_REGISTER_SET)
            instrRegs = regset_union(instrRegs, instr->arg1_set);
        if (instr->res_constraint == ARG_REGISTER_SET ||
            instr->res_constraint == ARG_BYTE_REGISTER_SET)
            instrRegs = regset_union(instrRegs, instr->res_set);
    } else if (argNumber == 1) {
        argConstraint = instr->arg1_constraint; /* Second argument */
        regSet = instr->arg1_set;
        if (instr->arg0_constraint == ARG_REGISTER_SET ||
            instr->arg0_constraint == ARG_BYTE_REGISTER_SET)
            instrRegs = regset_union(instrRegs, instr->arg0_set);
        if (instr->res_constraint == ARG_REGISTER_SET ||
            instr->res_constraint == ARG_BYTE_REGISTER_SET)
            instrRegs = regset_union(instrRegs, instr->res_set);
    } else {
        argConstraint = instr->res_constraint;  /* Result argument */
        regSet = instr->res_set;
        if (instr->arg0_constraint == ARG_REGISTER_SET ||
            instr->arg0_constraint == ARG_BYTE_REGISTER_SET)
            instrRegs = regset_union(instrRegs, instr->arg0_set);
        if (instr->arg1_constraint == ARG_REGISTER_SET ||
            instr->arg1_constraint == ARG_BYTE_REGISTER_SET)
            instrRegs = regset_union(instrRegs, instr->arg1_set);
    }

    preg = (-1);
    if (is_reg(arg)) {
        preg = arg->reg;
    } else if (is_mem(arg) || is_bytemem(arg)) {
        if (arg->left != NULL && arg->left->oper == RTX_REG) {
            preg = arg->left->reg;
        }
    }
    if (preg < 0)
        return ARG_RELOAD_NONE;

    hreg = pseudo_register_map[preg];

    if (argConstraint == ARG_REGISTER_SET || 
        argConstraint == ARG_BYTE_REGISTER ||
        argConstraint == ARG_BYTE_REGISTER_SET) { /* If not any register */
        possibleRegs = regSet;
    } else {
        possibleRegs = regset_subtract(REGISTER_FULL_SET, instrRegs);
    }
    possibleRegs = regset_subtract(possibleRegs, *reload_reg_set);
    assert(possibleRegs != 0);
    bestRegs = regset_intersect(possibleRegs, *free_reg_set);

    if (!reg_in_stack_slot(preg) &&
        regset_contains(possibleRegs, hreg)) {
        return ARG_RELOAD_NONE;         /* Nothing to do */
    }

    /* Reload is really needed */
    ret |= ARG_RELOAD_HREG;
    *reload_preg = preg;
    if (bestRegs != 0) {
        *reload_hreg = regset_get_first(bestRegs);
    } else {
        *reload_hreg = regset_get_first(possibleRegs);
    }

    if (!regset_contains(*free_reg_set, *reload_hreg)) {
        /* Need to save reload register */
        if (*free_reg_set != 0) {
            *saved_hreg = regset_get_first(*free_reg_set);
            *free_reg_set = regset_remove(*free_reg_set, *saved_hreg);
            ret |= ARG_RELOAD_SAVED_IN_HREG;
            generate_move_hregs(*reload_hreg, *saved_hreg);
        } else {
            ret |= ARG_RELOAD_SAVED_IN_STACK;
            generate_save_hreg(*reload_hreg);
        }
    }
    *reload_reg_set = regset_add(*reload_reg_set, *reload_hreg);
    if (argNumber != 2 ||       /* Not a result */
        !is_reg(arg)) {
        generate_move_pseudoreg_to_hreg(preg, *reload_hreg);
    }
    return ret;
}

void detect_basic_block(rtl_ptr line, int line_number) {
    rtl_ptr l = line;
    if (line->typ == RTL_INSTR) {
        /* Nothing to do, BB was defined at previous steps */
        return;
    }
    basic_block_begin = line_number;
    basic_block_end = line_number + 1;
    l = (rtl_ptr) line->link.next;
    while (l != (rtl_ptr) &rtl_function && l->typ == RTL_INSTR) {
        basic_block_end++;
        l = (rtl_ptr) l->link.next;
    }
}

static int write_global_definition(block_ptr n, void *arg1, long arg2) {
    name_ptr name = (name_ptr) n;
    type_ptr typ = &(name->typ);
    char number[64];
    memset(number, 0, 64);

    if ((typ->decl & DECL_FUN) && 
        !(name->modifiers & MOD_FUNCT_BODY_DEFINED)) {
        /* Extern function */
        genasm_write_line(
            gdef_extern_name,
            ARG_ESCAPE_CHARACTER,
            name->name, NULL, NULL
        );
    } else if (name->modifiers & MOD_EXTERN) {
        /* Extern variable */
        genasm_write_line(
            gdef_extern_name,
            ARG_ESCAPE_CHARACTER,
            name->name, NULL, NULL
        );
    } else if ((typ->base_type == BASETYPE_INT && typ->decl == DECL_VAR) ||
        typ->decl == DECL_PNT) {
        /* Simple int variable */
        genasm_write_line(
            gdef_int,
            ARG_ESCAPE_CHARACTER,
            name->name, NULL, NULL
        );
    } else if (typ->base_type == BASETYPE_CHAR && typ->decl == DECL_VAR) {
        /* Simple char variable */
        genasm_write_line(
            gdef_char,
            ARG_ESCAPE_CHARACTER,
            name->name, NULL, NULL
        );
    } else if ((typ->decl & DECL_ARR) && (
        typ->base_type == BASETYPE_INT ||
        (typ->base_type == BASETYPE_CHAR && (typ->decl & DECL_PNT))
    )) {
        /* int array */
        int size = 4 * typ->dimension;
        if (size <= 0) size = 4;
        sprintf(number, "%d", size);
        genasm_write_line(
            gdef_int_array,
            ARG_ESCAPE_CHARACTER,
            name->name, number, NULL
        );
    } else if ((typ->decl & DECL_ARR) &&
        typ->base_type == BASETYPE_CHAR &&
        !(typ->decl & DECL_PNT)
    ) {
        /* char array */
        int size = typ->dimension;
        if (size <= 0) size = 1;
        sprintf(number, "%d", size);
        genasm_write_line(
            gdef_char_array,
            ARG_ESCAPE_CHARACTER,
            name->name, number, NULL
        );
    }
    return 0;
}

void genasm_generate_global_definitions() {
    if (gdef_preambula != NULL)
        genasm_write_line(
            gdef_preambula,
            ARG_ESCAPE_CHARACTER,
            NULL, NULL, NULL
        );

    lst_loop(&global_table, &write_global_definition, NULL, 0);

    if (gdef_postambula != NULL)
        genasm_write_line(
            gdef_postambula,
            ARG_ESCAPE_CHARACTER,
            NULL, NULL, NULL
        );
}

static int write_string_definition(block_ptr s, void *arg1, long arg2) {
    char strbuf[1024];
    string_ptr str = (string_ptr) s;
    process_backslashs_in_string(str->str, strbuf, 1024);
    genasm_write_line(
        strdef_string,
        ARG_ESCAPE_CHARACTER,
        str->str_name, strbuf, NULL
    );
    return 0;
}

void genasm_generate_string_definitions() {
    if (strdef_preambula != NULL)
        genasm_write_line(
            strdef_preambula,
            ARG_ESCAPE_CHARACTER,
            NULL, NULL, NULL
        );

    lst_loop(&string_list, &write_string_definition, NULL, 0);

    if (strdef_postambula != NULL)
        genasm_write_line(
            strdef_postambula,
            ARG_ESCAPE_CHARACTER,
            NULL, NULL, NULL
        );
}

void process_backslashs_in_string(
    const char *src_str, char *dst_str, int maxlen
) {
    int i = 0;
    const char *p = src_str;
    while (*p != 0 && i < maxlen - 3) {
        if (*p == '\n') {
            dst_str[i++] = '\\'; dst_str[i++] = 'n';
            p++;
        } else if (*p == '\t') {
            dst_str[i++] = '\\'; dst_str[i++] = 't';
            p++;
        } else if (*p == '\f') {
            dst_str[i++] = '\\'; dst_str[i++] = 'f';
            p++;
        } else if (*p == '\r') {
            dst_str[i++] = '\\'; dst_str[i++] = 'r';
            p++;
        } else if (
            *p == '\\' &&
            *(p+1) != 'n' && *(p+1) != 't' && *(p+1) != 'f' &&
            *(p+1) != 'r' &&
            (!isdigit(*(p+1)) || *(p+1) >= '7')
        ) {
            dst_str[i++] = '\\'; dst_str[i++] = '\\';
            p++;
        } else if (
            (*p == '\"' || *p == '\'') &&
            (p == src_str || *(p-1) != '\\')
        ) {
            dst_str[i++] = '\\'; dst_str[i++] = *p;
            p++;
        } else {
            dst_str[i++] = *p++;
        }
    }
    dst_str[i] = 0;
}
