/*
 * Generation of Assembler code based on RTL representation
 */
#ifndef GENASM_H
#define GENASM_H

#include <stdio.h>

#include "rtl.h"

extern char output_file_name[];
extern FILE *output_file;

void genasm_init();
void genasm_terminate();
void genasm_write_file_preambula();
void genasm_write_file_postambula();

void genasm_generate_function();
void genasm_write_function_beg();
void genasm_write_function_end();
void genasm_write_text(
    const char * const lines[], /* Sequence of lines to write */
    int escape_character,       /* Character marking a place to insert */
    const char *insert0,        /* First pattern to insert */
    const char *insert1,        /* Second pattern to insert */
    const char *insert2         /* Third pattern to insert */
);
void genasm_write_line(
    const char *line,           /* Lines to write */
    int escape_character,       /* Character marking a place to insert */
    const char *insert0,        /* First pattern to insert */
    const char *insert1,        /* Second pattern to insert */
    const char *insert2         /* Third pattern to insert */
);
void genasm_substitute(
    char *line,
    int escape_character,       /* Character marking a place to insert */
    const char *insert0,        /* First pattern to insert */
    const char *insert1,        /* Second pattern to insert */
    const char *insert2         /* Third pattern to insert */
);

void genasm_allocate_registers();       /* First pass: register allocation */
void genasm_write_function_body();      /* Second pass: assembler output */

/* Register names */
extern const char * const REGISTER_NAME[];
extern const char * const BYTE_REGISTER_NAME[]; /* Low byte of register */

extern const int REGISTER_ALLOCATION_ORDER[];

/* The following constant must be defined in target machine specification */
/* (for example, we use I80386): */
/*...
#define NUM_REGISTERS           8
#define NUM_BYTE_REGISTERS      4
#define NUM_REGISTERS_ALLOCATE  6
#define REGISTER_FULL_SET       0x3f
#define SAVED_REGISTERS         (EBX_BIT | EDI_BIT | ESI_BIT)
#define BYTE_REGISTERS_SET      (EAX_BIT | EBX_BIT | ECX_BIT | EDX_BIT)
#define RETURN_REGISTERS        (EAX_BIT | EDX_BIT)
#define RETURN_REGISTER         EAX
...*/

/* Convertion register_bit -> register_index */
int register_index(int reg_bit);
int register_bit(int reg_idx);

/* Pseudo register usage */
extern int *pseudo_reg_assignment;
extern int *pseudo_reg_first_usage;
extern int *pseudo_reg_last_usage;
extern int pseudo_reg_array_size;

void scan_register_usage();

/* Hard register corresponding to pseudo reg */
extern int *pseudo_register_map; 
extern int *pseudo_register_stack_slot;

void release_pseudoreg(int pseudoreg);

/* Values of pseudo_register_map */
#define PSEUDO_REGISTER_UNMAPPED        (-1)
#define PSEUDO_REGISTER_IN_STACK_SLOT   NUM_REGISTERS
/* otherwise, it is a number of hard register */

/* Stack slots management */
extern int stack_32_slots;          /* Map of first 32 stack slots */
extern int stack_additional_space;  /* In words, including first 32 slots */
extern int stack_max_additional_space; /* max value */
extern int stack_slots_offset;  /* Offset from FP to bottom of stack slots */
void allocate_stack_slot(int pseudoreg);
void release_stack_slot(int slot);

/* Additional stack space for saving hard registers, above stack slots */
extern int stack_save_hardregs_depth;
extern int stack_save_hardregs_maxdepth;

/* Basic block */
extern int basic_block_begin; /* Line number of "goto" or label line */
extern int basic_block_end;
void detect_basic_block(rtl_ptr line, int line_number);

/* Work with register sets */
int regset_add(int set, int reg);
int regset_remove(int set, int reg);
int regset_contains(int set, int reg);
int regset_erase(int set);
int regset_subtract(int set, int set1);
int regset_union(int set, int set1);
int regset_intersect(int set, int set1);
int regset_loop(
    int set,
    int (*action)(int reg, void *arg1, long arg2),
    void *arg1, long arg2
);
int regset_reverse_loop(
    int set,
    int (*action)(int reg, void *arg1, long arg2),
    void *arg1, long arg2
);
int regset_complement(int set);
int regset_get_first(int set);
int regset_get_next(int set, int element);
int regset_get_last(int set);
int regset_get_prev(int set, int element);

/* Registers clobbered in function body */
extern int clobbered_registers;

/* Current set of allocated registers */
extern int allocated_registers;

extern int genasm_pass; /* Pass counter */

/* Definition of instruction set */
typedef struct instr_def_tag {
    int rtx_oper;               /* RTX operation */
    int rtx_num_args;           /* Number of args of RTX oper., incuding res */
    int rtx_arg0_constraint;    /* Type of first arg. of RTX expression */
    int rtx_arg0_value;         /* Used for constant types */
    int rtx_arg1_constraint;    /* Type of second arg. of RTX expression */
    int rtx_arg1_value;         /* Used for constant types */
    int rtx_res_constraint;     /* Type of RTX result */
    int rtx_res_value;          /* Used when result is in hard register */

    int num_args;               /* Number of args of instr. including res */
    int arg0_constraint;        /* Type of first argument */
    int arg0_set;               /* Set of registers to be used for arg0 */
    int arg1_constraint;        /* Type of second argument */
    int arg1_set;               /* Set of registers to be used for arg1 */
    int res_constraint;         /* Type of result */
    int res_set;                /* Set of registers to be used for res */
    int scratch_regs;           /* Set of temporary registers */
    int res_in_arg;             /* Result is in argument n */
    const char *templ;          /* Instruction template */
} instr_def, *instr_def_ptr;
typedef const instr_def *instr_def_cptr;

/* RTX argument constraints */
#define ARG_RTX_NONE            0
#define ARG_RTX_ANY             1   /* Any means int argument, not byte! */
#define ARG_RTX_REGISTER        2
#define ARG_RTX_MEMORY          3
#define ARG_RTX_BYTE_MEMORY     4
#define ARG_RTX_ANY_CONST       5
#define ARG_RTX_CONST           6   /* Some fixed constant (e.g 2, 4, 8) */
#define ARG_RTX_CC0             7
#define ARG_RTX_HARDREG         8
#define ARG_RTX_MEMREG_PREDEC   9   /* For push instruction */
#define ARG_RTX_MEMREG_POSTINC  10  /* For pop instruction (in fact, not used) */
#define ARG_RTX_NAME            11  /* Symbolic name */

/* Instruction argument constraints */
#define ARG_NONE                0
#define ARG_ANY                 1
#define ARG_REGISTER            2
#define ARG_REGISTER_SET        3
#define ARG_BYTE_REGISTER       4
#define ARG_BYTE_REGISTER_SET   5
#define ARG_MEMORY              6
#define ARG_BYTE_MEMORY         7
#define ARG_CONST               8       /* Intermediate const. $123 */
#define ARG_NAME                9       /* Intermediate name   $abc */
#define ARG_PLAIN_CONST         10      /* Plain const. 123 */
#define ARG_PLAIN_NAME          11      /* Plain const. abc */

/* Possible values of "res_in_arg" field, used for 2-address instructions */
#define RES_SEPARATE            0
#define RES_IN_ARG0             1
#define RES_IN_ARG1             2

#define ARG_ESCAPE_CHARACTER ';'    /* Semicolon is not used in any instr. */
#define ARG0_TEMPLATE        '0'
#define ARG1_TEMPLATE        '1'
#define RES_TEMPLATE         '2'

/* Set of instructions */
extern const instr_def instr_set[];

/* Arguments of instructions */
extern const char * const arg_const;            /* $10 */
extern const char * const arg_name;             /* $name */
extern const char * const arg_global_name;      /* $name, $_name in CYGWIN */
extern const char * const arg_mem_name;         /* name */
extern const char * const arg_mem_global_name;  /* name, _name in CYGWIN */
extern const char * const arg_mem_reg;          /* (%esi) */
extern const char * const arg_mem_reg_const;    /* 12(%ebp) */

/* Templates of some fixed instruction (pointers to instr_set[]) */
extern const instr_def_ptr INSTR_MOV_REG_ANY;
extern const instr_def_ptr INSTR_MOV_ANY_REG;
extern const instr_def_ptr INSTR_MOV_REG_BYTE;
extern const instr_def_ptr INSTR_MOV_BYTE_REG;
extern const instr_def_ptr INSTR_PUSH;
extern const instr_def_ptr INSTR_POP;
extern const instr_def_ptr INSTR_CALL;

extern const char * const file_preambula[];
extern const char * const file_postambula[];
extern const char * const function_preambula[];
extern const char * const stack_advance_instruction;
extern const char * const function_postambula[];
extern const char * const label_line;
extern const char * const goto_instruction;
extern const char * const branch_instruction;
extern const char * const asm_condition_names[];

/* Templates for global definitions */
extern const char * const gdef_preambula;
extern const char * const gdef_postambula;
extern const char * const gdef_int;
extern const char * const gdef_int_array;
extern const char * const gdef_char;
extern const char * const gdef_char_array;
extern const char * const gdef_extern_name;

/* Templates for string definitions */
extern const char * const strdef_preambula;
extern const char * const strdef_postambula;
extern const char * const strdef_string;

void genasm_write_goto(int label);
void genasm_write_branch(int label, int condition);
const char *genasm_cond_name(int condition);
void genasm_write_push(int reg);
void genasm_write_pop(int reg);

void genasm_assign_registers(rtl_ptr line, int line_number);
void genasm_write_instr(rtl_ptr line, int line_number);

void genasm_generate_global_definitions();
void genasm_generate_string_definitions();
void process_backslashs_in_string(
    const char *src_str, char *dst_str, int maxlen
);

#endif /* GENASM_H */
