#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "genrtl.h"
#include "rtl.h"
#include "nametab.h"
#include "scanner.h"
#include "strpool.h"
#include "comp.h"

static int label = 100;
static int pseudoreg = MIN_PSEUDOREG;

int gen_expr(tree_ptr e, int level) { /* Generate code for ariphm. expr. */
    int op;
    int res = 0; /* Pseudo register for result */

    assert(e != NULL);
    /*... print_tree(e); ...*/

    if (e->modifiers & MOD_BOOL)
        return gen_l2int(e);    /* Evaluate logical expression */

    op = e->oper;
    switch (op) {
    case NAME:
        res = gen_readvar(e->name); break;
    case ASG:
        res = gen_assign(e, level); break;
    case ICON:
        res = gen_getregister();
        gen_setreg(res, e->int_value); break;
    case STRING:
        res = gen_getregister();
        gen_setregstr(res, e->string_value); break;
    case PLUS:
    case MINUS:
    case MUL:
    case AND:
    case OR:
    case XOR:
    case DIV:
    case MOD:
    case LS:
    case RS:
        res = gen_binop(e); break;
    case LAND:
    case LOR:
    case LNOT:
    case LT:
    case LE:
    case GT:
    case GE:
    case EQ:
    case NE:
    case LTU:
    case LEU:
    case GTU:
    case GEU:
        assert(0); break;       /* Logical expression */
    case ICR:
    case DECR:
        res = gen_incrop(e, level); break;
    case PNT:
        res = gen_getregister();
        gen_setregrtx(res, gen_rtxderef(e));
        break;
    case FUN:
        res = gen_funcall(e, level); break;
    case NOT:
    case UNMINUS:
        res = gen_unop(e); break;
    case AMP:
        res = gen_getregister();
        gen_setregrtx(res, gen_rtxaddr(e->left));
        break;
    case LST: ;
    }
    return res;
}

void gen_lexpr(                     /* Generate code for logical expr.   */
    tree_ptr e,                     /* expression tree                   */
    int trueLabel, int falseLabel   /* trueLabel == 0 || falseLabel == 0 */
) {
    assert(trueLabel == 0 || falseLabel == 0);

    if (!(e->modifiers & MOD_BOOL)) {
        /* Ariphm. expression */
        int cond = EQ;
        if (falseLabel == 0)
            cond = NE;
        gen_tst(e);
        gen_condgoto(cond, trueLabel + falseLabel);
        return;
    }

    /* Logical expression */
    switch (e->oper) {
        case LNOT:
            gen_lexpr(e->left, falseLabel, trueLabel); break;
        case LAND:
            if (falseLabel != 0) {
                gen_lexpr(e->left, 0, falseLabel);
                gen_lexpr(e->right, 0, falseLabel);
            } else {
                int lab = gen_getlabel();
                gen_lexpr(e->left, 0, lab);
                gen_lexpr(e->right, trueLabel, 0);
                gen_label(lab);
            }
            break;
        case LOR:
            if (trueLabel != 0) {
                gen_lexpr(e->left, trueLabel, 0);
                gen_lexpr(e->right, trueLabel, 0);
            } else {
                int lab = gen_getlabel();
                gen_lexpr(e->left, lab, 0);
                gen_lexpr(e->right, 0, falseLabel);
                gen_label(lab);
            }
            break;
        case LT:
        case LE:
        case GT:
        case GE:
        case EQ:
        case NE:
        case LTU:
        case LEU:
        case GTU:
        case GEU:
            gen_cmp(e->left, e->right);
            if (trueLabel != 0)
                gen_condgoto(e->oper, trueLabel);
            else
                gen_condgoto(inverse_cond(e->oper), falseLabel);
            break;
    }
}

int gen_l2int(tree_ptr expr) { /* Evaluate logical expression */
    int lab = gen_getlabel();
    int reg = gen_getregister();

    gen_setreg(reg, 0);             /* res = FALSE           */
    gen_lexpr(expr, 0, lab);        /* if (!lexpr) goto lab; */
    gen_setreg(reg, 1);             /* res = TRUE            */
    add_rtl_label(lab);             /* lab: ;                */

    return reg;
}

int gen_getlabel() {                /* generate label */
    return label++;
}

int gen_getregister() {             /* generate pseudo-register */
    return pseudoreg++;
}

int gen_lastregister() {
    return pseudoreg - 1;
}

void gen_init_register_counter() {
    pseudoreg = MIN_PSEUDOREG;
}

void gen_start_function() {
    gen_init_register_counter();
}

void gen_label(int label) {
    add_rtl_label(label);
}

void gen_goto(int label) {
    add_rtl_goto(label);
}

void gen_condgoto(int cond, int label) {
    add_rtl_cond_goto(cond, label);
}

void gen_return(tree_ptr expr) {
    if (expr != NULL) {
        rtl_ptr l;
        rtx_ptr r;
        if (expr->oper == ICON) {
            r = gen_rtxconst(expr->int_value);
        } else {
            int res = gen_expr(expr, 1);  /* Calculate the return expr. */
            r = gen_rtxreg(res);
        }

        /* Copy return value to R0 */
        l = allocate_rtl_line();
        l->typ = RTL_INSTR;
        l->expr = allocate_rtx();
        l->expr->oper = RTX_SET;
        l->expr->left = gen_rtxhardreg(0);
        l->expr->right = r;
        add_rtl_line(l);
    }

    add_rtl_goto(function_end_label);
}

void gen_setreg(int reg, int val) {
    rtl_ptr l = allocate_rtl_line();
    l->typ = RTL_INSTR;
    l->expr = allocate_rtx();
    l->expr->oper = RTX_SET;
    l->expr->left = gen_rtxreg(reg);
    l->expr->right = gen_rtxconst(val);
    add_rtl_line(l);
}

void gen_setregrtx(int reg, rtx_ptr expr) {
    rtl_ptr l = allocate_rtl_line();
    l->typ = RTL_INSTR;
    l->expr = allocate_rtx();
    l->expr->oper = RTX_SET;
    l->expr->left = gen_rtxreg(reg);
    l->expr->right = expr;
    add_rtl_line(l);
}

void gen_sethardregrtx(int hreg, rtx_ptr expr) {
    rtl_ptr l = allocate_rtl_line();
    l->typ = RTL_INSTR;
    l->expr = allocate_rtx();
    l->expr->oper = RTX_SET;
    l->expr->left = gen_rtxhardreg(hreg);
    l->expr->right = expr;
    add_rtl_line(l);
}

void gen_setregstr(int reg, string_ptr str) {
    rtl_ptr l = allocate_rtl_line();
    l->typ = RTL_INSTR;
    l->expr = allocate_rtx();
    l->expr->oper = RTX_SET;
    l->expr->left = gen_rtxreg(reg);
    l->expr->right = gen_rtxstring(str);
    add_rtl_line(l);
}

int gen_readvar(name_ptr n) {          /* Read a simple variable */
    int dstreg = gen_getregister();
    rtl_ptr l = allocate_rtl_line();
    l->typ = RTL_INSTR;
    l->expr = allocate_rtx();
    l->expr->oper = RTX_SET;
    l->expr->left = gen_rtxreg(dstreg);
    l->expr->right = gen_rtxvar(n);
    add_rtl_line(l);
    return dstreg;
}

rtx_ptr gen_rtxreg(int pseudoreg) {
    rtx_ptr e = allocate_rtx();
    e->oper = RTX_REG;
    e->reg = pseudoreg;
    return e;
}

rtx_ptr gen_rtxhardreg(int hreg) {
    rtx_ptr e = allocate_rtx();
    e->oper = RTX_HARDREG;
    e->hard_reg = hreg;
    return e;
}

rtx_ptr gen_rtxcc0() {
    return gen_rtxhardreg(CC0);
}

rtx_ptr gen_rtxconst(int c) {
    rtx_ptr e = allocate_rtx();
    e->oper = RTX_ICON;
    e->int_value = c;
    return e;
}

rtx_ptr gen_rtxstring(string_ptr str) {
    rtx_ptr e = allocate_rtx();
    e->oper = RTX_NAME;
    e->string_value = str->str_name;
    e->int_value = str->str_id;
    return e;
}

rtx_ptr gen_rtxvar(name_ptr n) {        /* RTX for a simple variable */
    rtx_ptr e;
    rtx_ptr s = gen_rtxvaraddr(n);
    if (n->typ.decl & DECL_ARR)
        return s;               /* Array name is its address */
    e = allocate_rtx();
    e->oper = RTX_MEM;
    if (n->typ.base_type == BASETYPE_CHAR && !(n->typ.decl & DECL_PNT))
        e->modifiers |= RTX_MODBYTE;
    e->left = s;
    if ((n->modifiers & MOD_LOCAL) && !(n->modifiers & MOD_EXTERN)) {
        e->local_name = n;
    } else {
        e->global_name = n;
    }
    return e;
}

rtx_ptr gen_rtxvaraddr(name_ptr n) { /* RTX for an address of a variable */
    rtx_ptr s = allocate_rtx();
    if ((n->modifiers & MOD_LOCAL) && !(n->modifiers & MOD_EXTERN)) {
        /* Local variable */
        rtx_ptr s1 = allocate_rtx();
        rtx_ptr s2 = allocate_rtx();
        s->oper = RTX_PLUS;
        s->left = s1; s->right = s2;
        s1->oper = RTX_HARDREG; s1->hard_reg = FP;
        s2->oper = RTX_ICON; s2->int_value = n->offset;
    } else {
        /* Global variable */
        s->oper = RTX_NAME; s->string_value = n->name;
        s->modifiers |= RTX_MODGLOBAL;
    }
    return s;
}

int gen_assign(tree_ptr e, int level) {
    rtx_ptr rexpr;
    int res = 0;
    rtl_ptr l = allocate_rtl_line();

    assert(e->left != NULL && e->right != NULL);

    l->typ = RTL_INSTR;
    l->expr = allocate_rtx();
    l->expr->oper = RTX_SET;
    l->expr->left = gen_rtxlval(e->left);

    if (e->right->oper == ICON) {
        if (level == 0) {
            rexpr = gen_rtxconst(e->right->int_value);
        } else {
            res = gen_getregister();
            gen_setreg(res, e->right->int_value);
            rexpr = gen_rtxreg(res);
        }
    } else {
        res = gen_expr(e->right, level+1);
        rexpr = gen_rtxreg(res);
    }
    l->expr->right = rexpr;
    add_rtl_line(l);
    return res;
}

rtx_ptr gen_rtxderef(tree_ptr e) {      /* RTX for dereference */
    int r;
    rtx_ptr x;

    assert(e->oper == PNT);
    if ((r = gen_expr(e->left, 1)) < 0) return NULL;    /* Error */

    x = allocate_rtx();
    x->oper = RTX_MEM;
    if (e->modifiers & MOD_BYTE)
        x->modifiers |= RTX_MODBYTE;
    x->left = gen_rtxreg(r);
    return x;
}

rtx_ptr gen_rtxaddr(tree_ptr e) {       /* RTX for object address */
    int r;
    assert(e->modifiers & MOD_LVAL);
    if (e->oper == NAME)
        return gen_rtxvaraddr(e->name);
    assert(e->oper == PNT && e->left != NULL);
    if ((r = gen_expr(e->left, 1)) < 0) return NULL;    /* Error */
    return gen_rtxreg(r);
}

rtx_ptr gen_rtxlval(tree_ptr e) {       /* RTX for an object */
    assert(e->modifiers & MOD_LVAL);
    if (e->oper == NAME) {
        return gen_rtxvar(e->name);
    } else {
        assert(e->oper == PNT);
        return gen_rtxderef(e);
    }
}

int gen_binop(tree_ptr e) {
    int dstreg;
    int rtlop;
    rtx_ptr expr, exprl, exprr;

    assert(e->left != NULL && e->right != NULL);
    if (e->left->oper == ICON) {
        exprl = gen_rtxconst(e->left->int_value);
    } else {
        int regl = gen_expr(e->left, 1);
        exprl = gen_rtxreg(regl);
    }

    if (e->right->oper == ICON) {
        exprr = gen_rtxconst(e->right->int_value);
    } else {
        int regr = gen_expr(e->right, 1);
        exprr = gen_rtxreg(regr);
    }

    expr = allocate_rtx();
    switch (e->oper) {
        case PLUS: rtlop = RTX_PLUS; break;
        case MINUS: rtlop = RTX_MINUS; break;
        case MUL: rtlop = RTX_MUL; break;
        case AND: rtlop = RTX_AND; break;
        case OR: rtlop = RTX_OR; break;
        case XOR: rtlop = RTX_XOR; break;
        case DIV: rtlop = RTX_DIV; break;
        case MOD: rtlop = RTX_MOD; break;
        case LS: rtlop = RTX_LSHIFT; break;
        case RS: rtlop = RTX_RSHIFT; break;
        default: assert(0);
    }
    expr->oper = rtlop;
    if ((e->oper != PLUS && e->oper != MINUS) ||
        (e->left->pointer == 0 && e->right->pointer == 0)) {
        expr->left = exprl; expr->right = exprr;
    } else if (e->oper == PLUS) {
        if (e->left->pointer != 0) {
            assert(e->right->pointer == 0);
            expr->left = exprl;
            if (e->left->pointer == 1 && (
                e->left->base_type == BASETYPE_CHAR ||
                e->left->base_type == BASETYPE_VOID)) {
                expr->right = exprr;    /* Just a plain sum */
            } else {
                /* Scale the second operand by a factor 4 */
                int reg_offset = gen_getregister();
                rtx_ptr e = allocate_rtx();
                e->oper = RTX_MUL;
                e->left = exprr;
                e->right = gen_rtxconst(4);
                gen_setregrtx(reg_offset, e);

                /* Add it to the sum */
                expr->right = gen_rtxreg(reg_offset);
            }
        } else {
            assert(e->left->pointer == 0);
            assert(e->right->pointer != 0);
            /* Swap operands so that a pointer be first */
            expr->left = exprr;
            if (e->right->pointer == 1 && (
                e->right->base_type == BASETYPE_CHAR ||
                e->right->base_type == BASETYPE_VOID)) {
                expr->right = exprl;    /* Just a plain sum */
            } else {
                /* Scale by a factor 4 */
                int reg_offset = gen_getregister();
                rtx_ptr e = allocate_rtx();
                e->oper = RTX_MUL;
                e->left = exprl;
                e->right = gen_rtxconst(4);
                gen_setregrtx(reg_offset, e);

                /* Add it to the sum */
                expr->right = gen_rtxreg(reg_offset);
            }
        }
    } else if (e->oper == MINUS) {
        assert(e->left->pointer > 0 || e->right->pointer > 0);
        if (e->left->pointer > 0 && e->right->pointer > 0) {
            if (e->left->pointer == 1 && (
                e->left->base_type == BASETYPE_CHAR ||
                e->left->base_type == BASETYPE_VOID)) {
                expr->left = exprl;     /* Just a plain subtraction */
                expr->right = exprr;
            } else {
                /* Subtract at first */
                expr->left = exprl;
                expr->right = exprr;
                dstreg = gen_getregister();
                gen_setregrtx(dstreg, expr);

                /* Than divide difference by a factor 4 */
                expr = allocate_rtx();
                expr->oper = RTX_DIV;
                expr->left = gen_rtxreg(dstreg);
                expr->right = gen_rtxconst(4);
            }
        }
    }
    dstreg = gen_getregister();
    gen_setregrtx(dstreg, expr);
    return dstreg;
}

int gen_incrop(tree_ptr e, int level) {
    int rtlop;
    rtx_ptr object;
    rtl_ptr l;
    int increment;
    int reg0, reg1;

    assert(e->left != NULL && (e->left->modifiers & MOD_LVAL));

    rtlop = RTX_PLUS;
    if (e->oper == DECR)
        rtlop = RTX_MINUS;
    increment = 1;
    if (e->left->pointer == 1 && e->left->base_type != BASETYPE_CHAR)
        increment = 4;

    object = gen_rtxlval(e->left);      /* RTX for an object */
    reg0 = gen_getregister();
    gen_setregrtx(reg0, clone_rtx(object));

    /* Increment/decrement a value of object */
    reg1 = gen_getregister();
    l = allocate_rtl_line();
    l->typ = RTL_INSTR;
    l->expr = allocate_rtx();
    l->expr->oper = RTX_SET;
    l->expr->left = gen_rtxreg(reg1);
    l->expr->right = allocate_rtx();
    l->expr->right->oper = rtlop;       /* RTX_PLUS or RTX_MINUS */
    l->expr->right->left = gen_rtxreg(reg0);
    l->expr->right->right = gen_rtxconst(increment);
    add_rtl_line(l);

    /* Copy the changed value back to an object */
    l = allocate_rtl_line();
    l->typ = RTL_INSTR;
    l->expr = allocate_rtx();
    l->expr->oper = RTX_SET;
    l->expr->left = object;
    l->expr->right = gen_rtxreg(reg1);
    add_rtl_line(l);

    if (e->modifiers & MOD_POSTFIX)
        return reg0;
    else
        return reg1;
}

int gen_unop(tree_ptr e) {
    int dstreg, regl;
    int rtlop;
    rtx_ptr expr;
    int val;

    assert(e->left != NULL);
    assert(e->oper == UNMINUS || e->oper == NOT);

    if (e->left->oper == ICON) {
        val = e->left->int_value;
        if (e->oper == UNMINUS)
            val = (-val);
        else
            val = (~val);
        expr = gen_rtxconst(val);
    } else {
        rtlop = RTX_UNMINUS;
        if (e->oper == NOT)
            rtlop = RTX_NOT;
        expr = allocate_rtx();
        expr->oper = rtlop;

        regl = gen_expr(e->left, 1);
        expr->left = gen_rtxreg(regl);
    }
    dstreg = gen_getregister();
    gen_setregrtx(dstreg, expr);
    return dstreg;
}

int gen_cmp(tree_ptr e1, tree_ptr e2) {
    rtx_ptr expr, exprl, exprr;

    assert(e1 != NULL);
    if (e1->oper == ICON) {
        exprl = gen_rtxconst(e1->int_value);
    } else {
        int regl = gen_expr(e1, 1);
        exprl = gen_rtxreg(regl);
    }

    if (e2 == NULL) {
        /* TST operation */
        exprr = gen_rtxconst(0);
    } else if (e2->oper == ICON) {
        exprr = gen_rtxconst(e2->int_value);
    } else {
        int regr = gen_expr(e2, 1);
        exprr = gen_rtxreg(regr);
    }

    expr = allocate_rtx();
    expr->oper = RTX_CMP;
    expr->left = exprl; expr->right = exprr;

    gen_sethardregrtx(CC0, expr);
    return CC0;
}

int gen_tst(tree_ptr e) {
    return gen_cmp(e, NULL);
}

int gen_funcall(tree_ptr e, int level) { /* Function call */
    int res = 0;
    rtl_ptr l;
    tree_ptr lst;
    rtx_ptr rexpr;
    int numArgs = 0;

    assert(e != NULL);

    /* Process list of arguments */
    lst = e->left;
    while (lst != NULL) {
        /* Process argument */
        l = allocate_rtl_line();

        assert(lst->oper == LST);
        assert(lst->left != NULL);

        /* Push arg to stack */
        l->typ = RTL_INSTR;
        l->expr = allocate_rtx();
        l->expr->oper = RTX_SET;

        /* The left part of assignment */
        l->expr->left = allocate_rtx();
        l->expr->left->oper = RTX_MEM;
        l->expr->left->left = allocate_rtx();
        l->expr->left->left->oper = RTX_PREDEC;
        l->expr->left->left->left = gen_rtxhardreg(SP);

        /* The right part of assignment */
        if (lst->left->oper == ICON) {
            rexpr = gen_rtxconst(lst->left->int_value);
        } else {
            int arg = gen_expr(lst->left, level+1);
            rexpr = gen_rtxreg(arg);
        }
        l->expr->right = rexpr;
        add_rtl_line(l);
        numArgs++;

        /* To next argument */
        lst = lst->right;
    }

    /* Generate function call */
    /* gcc generates (call (mem:QI (symbol_ref:SI ("f"))) */
    /*                      ?????? I omitted it           */
    l = allocate_rtl_line();
    l->typ = RTL_INSTR;
    l->expr = allocate_rtx();
    l->expr->oper = RTX_SET;
    l->expr->left = gen_rtxhardreg(0);

    l->expr->right = allocate_rtx();
    l->expr->right->oper = RTX_CALL;
    l->expr->right->left = allocate_rtx();
    l->expr->right->left->oper = RTX_MEM;
    l->expr->right->left->left = allocate_rtx();
    l->expr->right->left->left->oper = RTX_NAME;
    l->expr->right->left->left->string_value = e->name->name;
    if (e->name->modifiers & (MOD_EXTERN | MOD_GLOBAL))
        l->expr->right->left->left->modifiers |= RTX_MODGLOBAL;
    add_rtl_line(l);

    /* Release stack */
    if (numArgs > 0) {
        l = allocate_rtl_line();
        l->typ = RTL_INSTR;
        l->expr = allocate_rtx();
        l->expr->oper = RTX_SET;
        l->expr->left = gen_rtxhardreg(SP);
        l->expr->right = allocate_rtx();
        l->expr->right->oper = RTX_PLUS;
        l->expr->right->left = gen_rtxhardreg(SP);
        l->expr->right->right = gen_rtxconst(numArgs * 4);
        add_rtl_line(l);
    }

    /* Copy return value from R0 to pseudoregister */
    if (level > 0) {
        l = allocate_rtl_line();
        l->typ = RTL_INSTR;
        l->expr = allocate_rtx();
        l->expr->oper = RTX_SET;
        res = gen_getregister();
        l->expr->left = gen_rtxreg(res);
        l->expr->right = gen_rtxhardreg(0);
        add_rtl_line(l);
    }
    return res;
}

int check_lvalue(tree_ptr expr) {
    return ((expr->modifiers & MOD_LVAL) != 0);
}
