#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "nametab.h"
#include "tree.h"
#include "scanner.h"
#include "comp.h"

static const char* tree_node_oper(tree_ptr t);

block_header tree_nodes_allocated; /* List of all tree nodes */

void init_trees() {
    lst_link(&tree_nodes_allocated, &tree_nodes_allocated);
}

void release_tree(tree_ptr t) {
    if (t == NULL)
        return;
    if (t->right != NULL)
        release_tree(t->right);
    if (t->left != NULL)
        release_tree(t->left);
    release_tree_node(t);
}

tree_ptr allocate_tree_node() {
    tree_ptr t = (tree_ptr) malloc(sizeof(tree));
    memset(t, 0, sizeof(tree));
    lst_addAfter(&tree_nodes_allocated, &(t->link));
    return t;
}

void release_tree_node(tree_ptr node) {
    if (node != NULL) {
        lst_remove(&(node->link));
        free(node);
    }
}

void release_all_trees() {
    block_ptr n = tree_nodes_allocated.next;
    if (compiler_flags & DUMP_TREES)
        fprintf(dump_file, "Releasing all trees\n");
    while (n != &tree_nodes_allocated) {
        lst_remove(n);
        free(n);
        n = tree_nodes_allocated.next;
    }
}

void print_tree(tree_ptr t) {
    const char *n = "undefined";
    if (t == NULL || !(compiler_flags & DUMP_TREES))
        return;

    fprintf(dump_file, "tree node ");

    if (t->oper == NAME || t->oper == FUN || t->oper == ARR) {
        if (t->name != NULL)
            n = t->name->name;
        fprintf(dump_file, "%p %s \"%s\" left=%p right=%p\n",
            t, tree_node_oper(t), n, t->left, t->right);
    } else if (t->oper == ICON) {
        fprintf(dump_file, "%p %s %d left=%p right=%p\n",
            t, tree_node_oper(t), t->int_value, t->left, t->right);
    } else if (t->oper == STRING) {
        fprintf(dump_file, "%p %s \"%s\" (%p) left=%p right=%p\n",
            t, tree_node_oper(t), t->string_value->str, t->string_value->str,
            t->left, t->right);
    } else {
        fprintf(dump_file, "%p %s left=%p right=%p\n",
            t, tree_node_oper(t), t->left, t->right);
    }

    if (t->left != NULL)
        print_tree(t->left);
    if (t->right != NULL)
        print_tree(t->right);
}

static const char* tree_node_oper(tree_ptr t) {
    const char *oper = "???";
    if (t == NULL)
        return oper;
    switch (t->oper) {
        case NAME: oper = "NAME"; break;
        case ICON: oper = "ICON"; break;
        case STRING: oper = "STRING"; break;
        case PLUS: oper = "PLUS"; break;
        case MINUS: oper = "MINUS"; break;
        case MUL: oper = "MUL"; break;
        case AND: oper = "AND"; break;
        case LAND: oper = "LAND"; break;
        case OR: oper = "OR"; break;
        case LOR: oper = "LOR"; break;
        case XOR: oper = "XOR"; break;
        case NOT: oper = "NOT"; break;
        case LNOT: oper = "LNOT"; break;
        case ASG: oper = "ASG"; break;
        case RELOP: oper = "RELOP"; break;
        case EQUOP: oper = "EQUOP"; break;
        case DIVOP: oper = "DIVOP"; break;
        case SHIFTOP: oper = "SHIFTOP"; break;
        case ICOP: oper = "ICOP"; break;
        case PNT: oper = "PNT"; break;
        case ARR: oper = "ARR"; break;
        case FUN: oper = "FUN"; break;
        case UNMINUS: oper = "UNMINUS"; break;
        case AMP: oper = "AMP"; break;
        case LST: oper = "LST"; break;

        case LT: oper = "LT"; break;
        case LE: oper = "LE"; break;
        case GT: oper = "GT"; break;
        case GE: oper = "GE"; break;
        case EQ: oper = "EQ"; break;
        case NE: oper = "NE"; break;

        case DIV: oper = "DIV"; break;
        case MOD: oper = "MOD"; break;
        case LS: oper = "LS"; break;
        case RS: oper = "RS"; break;
        case ICR: oper = "ICR"; break;
        case DECR: oper = "DECR"; break;

        /* Unsigned comparings */
        case LTU: oper = "LTU"; break;
        case LEU: oper = "LEU"; break;
        case GTU: oper = "GTU"; break;
        case GEU: oper = "GEU"; break;
    };
    return oper;
}

int expr_type(tree_ptr e) {  /* Returns 0, if correct, (-1), if error */
    int res = 0;

    /* Calculate expr. types of sons */
    if (e->left != NULL && expr_type(e->left) < 0) res = (-1);
    if (e->right != NULL && expr_type(e->right) < 0) res = (-1);

    switch (e->oper) {
        case NAME:
            if (e->name == NULL) {
                res = (-1); break;      /* Undefined name */
            }
            e->base_type = e->name->typ.base_type;
            if (e->name->typ.decl & (DECL_PNT | DECL_ARR))
                e->pointer = 1;
            else if (e->name->typ.base_type == BASETYPE_CHAR)
                e->modifiers |= MOD_BYTE;
            if (!(e->name->typ.decl & (DECL_FUN | DECL_ARR)))
                e->modifiers |= MOD_LVAL;
            break;
        case ICON:
            e->base_type = BASETYPE_INT; break;
        case STRING:
            e->base_type = BASETYPE_CHAR; e->pointer = 1; break;
        case PLUS:
            if (e->left->pointer != 0 && e->right->pointer != 0) {
                error_message0("Invalid sum of pointers");
                res = (-1);
            } else if (e->left->pointer != 0 || e->right->pointer != 0) {
                if (e->left->pointer != 0) {
                    e->pointer = e->left->pointer;
                    e->base_type = e->left->base_type;
                } else {
                    e->pointer = e->right->pointer;
                    e->base_type = e->right->base_type;
                }
            } else if (
                e->left->base_type == BASETYPE_VOID ||
                e->right->base_type == BASETYPE_VOID
            ) {
                error_message0("\"void\" is used as operand of binary \"+\"");
                res = (-1);
            } else if (
                e->left->base_type == BASETYPE_INT ||
                e->right->base_type == BASETYPE_INT
            ) {
                e->base_type = BASETYPE_INT;
            } else {
                e->base_type = BASETYPE_CHAR;
            }
            break;
        case MINUS:
            if (e->left->pointer != 0 && e->right->pointer != 0) {
                e->base_type = BASETYPE_INT;
                e->pointer = 0;
            } else if (e->right->pointer != 0) {
                error_message0("Illegal expression with pointer");
                res = (-1);
            } else if (e->left->pointer != 0) {
                e->pointer = e->left->pointer;
                e->base_type = e->left->base_type;
            } else if (
                e->left->base_type == BASETYPE_VOID ||
                e->right->base_type == BASETYPE_VOID
            ) {
                error_message0("\"void\" is used as operand of binary \"-\"");
                res = (-1);
            } else if (
                e->left->base_type == BASETYPE_INT ||
                e->right->base_type == BASETYPE_INT
            ) {
                e->base_type = BASETYPE_INT;
            } else {
                e->base_type = BASETYPE_CHAR;
            }
            break;
        case MUL:
        case DIVOP:
        case AND:
        case OR:
        case XOR:
        case SHIFTOP:
            if (e->left->pointer != 0 || e->right->pointer != 0) {
                error_message0("Illegal expression with pointer");
                res = (-1);
            } else if (
                e->left->base_type == BASETYPE_VOID ||
                e->right->base_type == BASETYPE_VOID
            ) {
                error_message0("\"void\" is used as operand of expression");
                res = (-1);
            } else if (
                e->left->base_type == BASETYPE_INT ||
                e->right->base_type == BASETYPE_INT
            ) {
                e->base_type = BASETYPE_INT;
            } else {
                e->base_type = BASETYPE_CHAR;
            }
            break;
        case LAND:
        case LOR:
        case LNOT:
        case EQ:
        case NE:
        case GT:
        case GE:
        case LT:
        case LE:
            e->base_type = BASETYPE_INT;
            e->modifiers |= MOD_BOOL;
            if (
                (e->oper == GT || e->oper == GE ||
                e->oper == LT || e->oper == LE) &&
                (e->left != NULL && e->right != NULL &&
                (e->left->pointer != 0 || e->right->pointer != 0))
            ) {
                /* Change a condition to unsigned */
                switch (e->oper) {
                    case GT: e->oper = GTU; break;
                    case GE: e->oper = GEU; break;
                    case LT: e->oper = LTU; break;
                    case LE: e->oper = LEU; break;
                }
            }
            break;
        case ASG:
            if (!(e->left->modifiers & MOD_LVAL)) {
                error_message0("Left side of assignment is not L-value");
                res = (-1);
            } else {
                e->base_type = e->left->base_type;
                e->pointer = e->left->pointer;
                e->modifiers = e->left->modifiers;
            }
            break;
        case ICR:
        case DECR:
            if (!(e->left->modifiers & MOD_LVAL)) {
                error_message0(
                    "The argument of ++/-- operation is not L-value");
                res = (-1);
            } else {
                e->base_type = e->left->base_type;
                e->pointer = e->left->pointer;
                e->modifiers &= ~MOD_LVAL;
            }
            break;
        case PNT:
            if (e->left->base_type == BASETYPE_VOID &&
                e->left->pointer == 1) {
                error_message0(
                    "Illegal dereference of \"void\" pointer");
                res = (-1);
            }
            if (e->left->pointer == 0) {
                error_message0(
                    "The agument of the \"*\" operation is not a pointer");
                res = (-1);
            }
            e->base_type = e->left->base_type;
            if (e->left->pointer > 0)
                e->pointer = e->left->pointer - 1;
            e->modifiers |= MOD_LVAL;
            if (e->base_type == BASETYPE_CHAR && e->pointer == 0)
                e->modifiers |= MOD_BYTE;
            break;
        case ARR: assert(0); break; /* Not used any more */
        case FUN:
            if (e->name == NULL) {
                e->base_type = BASETYPE_INT;
            } else {
                e->base_type = e->name->typ.base_type;
                if (e->name->typ.decl & DECL_PNT)
                    e->pointer = 1;
                else if (e->name->typ.base_type == BASETYPE_CHAR)
                    e->modifiers |= MOD_BYTE;
            }
            break;
        case UNMINUS:
        case NOT:
            if (e->left->pointer != 0) {
                error_message0("Illegal expression with pointer");
                res = (-1);
            } else {
                e->base_type = e->left->base_type;
                e->pointer = e->left->pointer;
                e->modifiers = (e->left->modifiers & ~MOD_LVAL);
            }
            break;
        case AMP:
            if (!(e->left->modifiers & MOD_LVAL)) {
                error_message0(
                    "The argument of the \"&\" operation is not L-value");
                res = (-1);
            } else {
                e->base_type = e->left->base_type;
                e->pointer = e->left->pointer + 1;
                e->modifiers = (e->left->modifiers & ~MOD_LVAL);
            }
            break;
        case LST: 
            e->base_type = e->left->base_type;
            e->pointer = e->left->pointer;
            e->modifiers = e->left->modifiers;
            break;
    }
    if (res < 0)
        e->base_type = (-1);
    return res;
}
